from __future__ import annotations

import copy
import importlib
import os
import sys
import types
from collections import OrderedDict

if sys.version_info >= (3, 11):
    # Python 3.11, see https://docs.python.org/3/whatsnew/3.11.html#enum
    from enum import ReprEnum as Enum
else:
    from enum import Enum

import typing
from collections import OrderedDict as OrderedDictType
from collections.abc import Callable
from typing import Any
from typing import Generic
from typing import TypeVar

import pwndbg
import pwndbg.aglib.heap
import pwndbg.aglib.heap.heap
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.aglib.symbol
import pwndbg.aglib.tls
import pwndbg.aglib.typeinfo
import pwndbg.aglib.vmmap
import pwndbg.chain
import pwndbg.color.memory as mem_color
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.memory
import pwndbg.libc
import pwndbg.libc.glibc
from pwndbg.color import message
from pwndbg.lib import SymbolNotRecoveredError

PREV_INUSE = 1
IS_MMAPPED = 2
NON_MAIN_ARENA = 4
SIZE_BITS = PREV_INUSE | IS_MMAPPED | NON_MAIN_ARENA
NONCONTIGUOUS_BIT = 2

NBINS = 128
NSMALLBINS = 64

# The `pwndbg.aglib.heap.structs` module is only imported at runtime when
# the heap heuristics are used in `HeuristicHeap.struct_module` and
# uses runtime information to select the correct structs.
# Only import it globally during static type checking.
if typing.TYPE_CHECKING:
    import pwndbg.aglib.heap.structs

    TheType = TypeVar("TheType", pwndbg.dbg_mod.Type, type[pwndbg.aglib.heap.structs.CStruct2GDB])
    TheValue = TypeVar("TheValue", pwndbg.dbg_mod.Value, pwndbg.aglib.heap.structs.CStruct2GDB)
else:
    TheType = TypeVar("TheType")
    TheValue = TypeVar("TheValue")

# We defer the initialization of HEAP_MAX_SIZE, as it needs the value of
# `pwndbg.aglib.arch.ptrsize`, which might not always be available at this
# point. See `heap_for_ptr`.
HEAP_MAX_SIZE: int = None

NBINS = 128
BINMAPSIZE = 4
NFASTBINS = 10
NSMALLBINS = 64


# Note that we must inherit from `str` before `Enum`: https://stackoverflow.com/a/58608362/803801
class BinType(str, Enum):
    TCACHE = "tcachebins"
    FAST = "fastbins"
    SMALL = "smallbins"
    LARGE = "largebins"
    UNSORTED = "unsortedbin"
    NOT_IN_BIN = "not_in_bin"

    def valid_fields(self) -> list[str]:
        if self in [BinType.FAST, BinType.TCACHE]:
            return ["fd"]
        if self in [BinType.SMALL, BinType.UNSORTED]:
            return ["fd", "bk"]
        if self == BinType.LARGE:
            return ["fd", "bk", "fd_nextsize", "bk_nextsize"]
        # BinType.NOT_IN_BIN
        return []


class Bin:
    def __init__(
        self,
        fd_chain: list[int],
        bk_chain: list[int] | None = None,
        count: int | None = None,
        is_corrupted: bool = False,
    ) -> None:
        self.fd_chain = fd_chain
        self.bk_chain = bk_chain
        self.count = count
        self.is_corrupted = is_corrupted

    def contains_chunk(self, chunk: int) -> bool:
        return chunk in self.fd_chain

    @staticmethod
    def size_to_display_name(size: int | str) -> str:
        if isinstance(size, str) and size == "all":
            return size

        assert isinstance(size, int)

        return hex(size)


class Bins:
    def __init__(self, bin_type: BinType) -> None:
        self.bins: OrderedDictType[int | str, Bin] = OrderedDict()
        self.bin_type = bin_type

    # TODO: There's a bunch of bin-specific logic in here, maybe we should
    # subclass and put that logic in there
    def contains_chunk(self, size: int, chunk: int):
        # TODO: It will be the same thing, but it would be better if we used
        # pwndbg.aglib.heap.current.size_sz. I think each bin should already have a
        # reference to the allocator and shouldn't need to access the `current`
        # variable
        ptr_size = pwndbg.aglib.arch.ptrsize

        if self.bin_type == BinType.UNSORTED:
            # The unsorted bin only has one bin called 'all'

            # Handle this case here, so we don't assign a str to an int-type variable
            if "all" in self.bins:
                return self.bins["all"].contains_chunk(chunk)
            return False
        if self.bin_type == BinType.LARGE:
            # All the other bins (other than unsorted) store chunks of the same
            # size in a bin, so we can use the size directly. But the largebin
            # stores a range of sizes, so we need to compute which bucket this
            # chunk falls into

            # TODO: Refactor this, the bin should know how to calculate
            # largebin_index without calling into the allocator
            assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
            size = pwndbg.aglib.heap.current.largebin_index(size) - NSMALLBINS

        elif self.bin_type == BinType.TCACHE:
            # Unlike fastbins, tcache bins don't store the chunk address in the
            # bins, they store the address of the fd pointer, so we need to
            # search for that address in the tcache bin instead

            # TODO: Can we use chunk_key_offset?
            chunk += ptr_size * 2

        if size in self.bins:
            return self.bins[size].contains_chunk(chunk)

        return False


def heap_for_ptr(ptr: int) -> int:
    """Round a pointer to a chunk down to find its corresponding heap_info
    struct, the pointer must point inside a heap which does not belong to
    the main arena.
    """
    # See https://sourceware.org/git/?p=glibc.git;a=blob;f=malloc/arena.c;h=37183cfb6ab5d0735cc82759626670aff3832cd0;hb=086ee48eaeaba871a2300daf85469671cc14c7e9#l30
    # and https://sourceware.org/git/?p=glibc.git;a=blob;f=malloc/malloc.c;h=f8e7250f70f6f26b0acb5901bcc4f6e39a8a52b2;hb=086ee48eaeaba871a2300daf85469671cc14c7e9#l869
    # 1 Mb (x86) or 64 Mb (x64)
    global HEAP_MAX_SIZE
    if HEAP_MAX_SIZE is None:
        HEAP_MAX_SIZE = 1024 * 1024 if pwndbg.aglib.arch.ptrsize == 4 else 2 * 4 * 1024 * 1024 * 8

    return ptr & ~(HEAP_MAX_SIZE - 1)


class ChunkField(int, Enum):
    PREV_SIZE = 1
    SIZE = 2
    FD = 3
    BK = 4
    FD_NEXTSIZE = 5
    BK_NEXTSIZE = 6


def fetch_chunk_metadata(address: int, include_only_fields: set[ChunkField] | None = None):
    prev_size_field_name = pwndbg.aglib.memory.resolve_renamed_struct_field(
        "malloc_chunk", {"prev_size", "mchunk_prev_size"}
    )
    size_field_name = pwndbg.aglib.memory.resolve_renamed_struct_field(
        "malloc_chunk", {"size", "mchunk_size"}
    )

    if include_only_fields is None:
        fetched_struct = pwndbg.aglib.memory.fetch_struct_as_dictionary("malloc_chunk", address)
    else:
        requested_fields: set[str] = set()

        for field in include_only_fields:
            if field is ChunkField.PREV_SIZE:
                requested_fields.add(prev_size_field_name)
            elif field is ChunkField.SIZE:
                requested_fields.add(size_field_name)
            elif field is ChunkField.FD:
                requested_fields.add("fd")
            elif field is ChunkField.BK:
                requested_fields.add("bk")
            elif field is ChunkField.FD_NEXTSIZE:
                requested_fields.add("fd_nextsize")
            elif field is ChunkField.BK_NEXTSIZE:
                requested_fields.add("bk_nextsize")

        fetched_struct = pwndbg.aglib.memory.fetch_struct_as_dictionary(
            "malloc_chunk", address, include_only_fields=requested_fields
        )

    normalized_struct = {}
    for field in fetched_struct:
        if field == prev_size_field_name:
            normalized_struct[ChunkField.PREV_SIZE] = fetched_struct[prev_size_field_name]
        elif field == size_field_name:
            normalized_struct[ChunkField.SIZE] = fetched_struct[size_field_name]
        elif field == "fd":
            normalized_struct[ChunkField.FD] = fetched_struct["fd"]
        elif field == "bk":
            normalized_struct[ChunkField.BK] = fetched_struct["bk"]
        elif field == "fd_nextsize":
            normalized_struct[ChunkField.FD_NEXTSIZE] = fetched_struct["fd_nextsize"]
        elif field == "bk_nextsize":
            normalized_struct[ChunkField.BK_NEXTSIZE] = fetched_struct["bk_nextsize"]

    return normalized_struct


class Chunk:
    __slots__ = (
        "_gdbValue",
        "address",
        "_prev_size",
        "_size",
        "_real_size",
        "_flags",
        "_non_main_arena",
        "_is_mmapped",
        "_prev_inuse",
        "_fd",
        "_bk",
        "_fd_nextsize",
        "_bk_nextsize",
        "_heap",
        "_arena",
        "_is_top_chunk",
    )

    def __init__(self, addr: int, heap: Heap | None = None, arena: Arena | None = None) -> None:
        assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
        assert pwndbg.aglib.heap.current.malloc_chunk is not None
        if isinstance(pwndbg.aglib.heap.current.malloc_chunk, pwndbg.dbg_mod.Type):
            self._gdbValue = pwndbg.aglib.memory.get_typed_pointer_value(
                pwndbg.aglib.heap.current.malloc_chunk, addr
            )
        else:
            self._gdbValue = pwndbg.aglib.heap.current.malloc_chunk(addr)
        self.address = int(self._gdbValue.address)
        self._prev_size: int | None = None
        self._size: int | None = None
        self._real_size: int | None = None
        self._flags: dict[str, bool] | None = None
        self._non_main_arena: bool | None = None
        self._is_mmapped: bool | None = None
        self._prev_inuse: bool | None = None
        self._fd = None
        self._bk = None
        self._fd_nextsize = None
        self._bk_nextsize = None
        self._heap = heap
        self._arena = arena
        self._is_top_chunk: bool | None = None

    # Some chunk fields were renamed in GLIBC 2.25 master branch.
    def __match_renamed_field(self, field: str):
        field_renames = {
            "size": ["size", "mchunk_size"],
            "prev_size": ["prev_size", "mchunk_prev_size"],
        }

        for field_name in field_renames[field]:
            if self._gdbValue.type.has_field(field_name):
                return field_name

        raise ValueError(f"Chunk field name did not match any of {field_renames[field]}.")

    @property
    def prev_size(self) -> int | None:
        if self._prev_size is None:
            try:
                self._prev_size = int(self._gdbValue[self.__match_renamed_field("prev_size")])
            except pwndbg.dbg_mod.Error:
                pass

        return self._prev_size

    @property
    def size(self) -> int | None:
        if self._size is None:
            try:
                self._size = int(self._gdbValue[self.__match_renamed_field("size")])
            except pwndbg.dbg_mod.Error:
                pass

        return self._size

    @property
    def real_size(self) -> int | None:
        if self._real_size is None:
            try:
                self._real_size = int(self._gdbValue[self.__match_renamed_field("size")]) & ~(
                    SIZE_BITS
                )
            except pwndbg.dbg_mod.Error:
                pass

        return self._real_size

    @property
    def flags(self) -> dict[str, bool] | None:
        if self._flags is None:
            if (
                self.size is not None
                and self.non_main_arena is not None
                and self.is_mmapped is not None
                and self.prev_inuse is not None
            ):
                self._flags = {
                    "non_main_arena": self.non_main_arena,
                    "is_mmapped": self.is_mmapped,
                    "prev_inuse": self.prev_inuse,
                }

        return self._flags

    @property
    def non_main_arena(self) -> bool | None:
        if self._non_main_arena is None:
            sz = self.size
            if sz is not None:
                self._non_main_arena = bool(sz & NON_MAIN_ARENA)

        return self._non_main_arena

    @property
    def is_mmapped(self) -> bool | None:
        if self._is_mmapped is None:
            sz = self.size
            if sz is not None:
                self._is_mmapped = bool(sz & IS_MMAPPED)

        return self._is_mmapped

    @property
    def prev_inuse(self) -> bool | None:
        if self._prev_inuse is None:
            sz = self.size
            if sz is not None:
                self._prev_inuse = bool(sz & PREV_INUSE)

        return self._prev_inuse

    @property
    def fd(self):
        if self._fd is None:
            try:
                self._fd = int(self._gdbValue["fd"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._fd

    @property
    def bk(self):
        if self._bk is None:
            try:
                self._bk = int(self._gdbValue["bk"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._bk

    @property
    def fd_nextsize(self):
        if self._fd_nextsize is None:
            try:
                self._fd_nextsize = int(self._gdbValue["fd_nextsize"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._fd_nextsize

    @property
    def bk_nextsize(self):
        if self._bk_nextsize is None:
            try:
                self._bk_nextsize = int(self._gdbValue["bk_nextsize"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._bk_nextsize

    @property
    def heap(self) -> Heap:
        if self._heap is None:
            self._heap = Heap(self.address)

        return self._heap

    @property
    def arena(self) -> Arena | None:
        if self._arena is None:
            self._arena = self.heap.arena

        return self._arena

    @property
    def is_top_chunk(self):
        if self._is_top_chunk is None:
            ar = self.arena
            if ar is not None and self.address == ar.top:
                self._is_top_chunk = True
            else:
                self._is_top_chunk = False

        return self._is_top_chunk

    def next_chunk(self) -> Chunk | None:
        if self.is_top_chunk:
            return None

        if self.real_size is None or self.real_size == 0:
            return None

        next = Chunk(self.address + self.real_size, arena=self.arena)
        if pwndbg.aglib.memory.is_readable_address(next.address):
            return next
        return None

    def __contains__(self, addr: int) -> bool:
        """
        This allow us to avoid extra constructions like 'if start_addr <= ptr < end_addr', etc.
        """
        size_field_address = int(self._gdbValue[self.__match_renamed_field("size")].address)
        start_address = size_field_address if self.prev_inuse else self.address

        next = self.next_chunk()
        # and this is handles chunk's last qword field, depending on prev_inuse bit
        if next is None:
            end_address = size_field_address + self.real_size
        else:
            next_size_field_address = int(
                next._gdbValue[self.__match_renamed_field("size")].address
            )
            end_address = next_size_field_address if next.prev_inuse else next.address

        return start_address <= addr < end_address


class Heap:
    __slots__ = (
        "_gdbValue",
        "arena",
        "_memory_region",
        "start",
        "end",
        "_prev",
        "first_chunk",
    )

    start: int
    end: int

    def __init__(self, addr: int, arena: Arena | None = None) -> None:
        """Build a Heap object given an address on that heap.
        Heap regions are treated differently depending on their arena:
        1) main_arena - uses the sbrk heap
        2) non-main arena - heap starts after its heap_info struct (and possibly an arena)
        3) non-contiguous main_arena - just a memory region
        4) no arena - for fake/mmapped chunks
        """
        allocator = pwndbg.aglib.heap.current
        assert isinstance(allocator, GlibcMemoryAllocator)
        main_arena = allocator.main_arena

        sbrk_region = allocator.get_sbrk_heap_region()
        if sbrk_region is not None and addr in sbrk_region:
            # Case 1; main_arena.
            self.arena = main_arena if arena is None else arena
            self._memory_region = sbrk_region
            self._gdbValue = None
        else:
            heap_region = allocator.get_region(addr)
            if heap_region is None:
                raise ValueError(f"Cannot build heap object on an unmapped address ({hex(addr)})")

            heap_info = allocator.get_heap(addr)
            ar_ptr = None
            if heap_info is not None:
                ar_ptr = int(heap_info["ar_ptr"])

            if ar_ptr is not None and ar_ptr in (ar.address for ar in allocator.arenas):
                # Case 2; non-main arena.
                self.arena = Arena(ar_ptr) if arena is None else arena
                start = heap_region.start + allocator.heap_info.sizeof
                if ar_ptr in heap_region:
                    start += pwndbg.lib.memory.align_up(
                        allocator.malloc_state.sizeof, allocator.malloc_alignment
                    )

                heap_region.memsz = heap_region.end - start
                heap_region.vaddr = start
                self._memory_region = heap_region
                self._gdbValue = heap_info
            elif main_arena.non_contiguous:
                # Case 3; non-contiguous main_arena.
                self.arena = main_arena if arena is None else arena
                self._memory_region = heap_region
                self._gdbValue = None
            else:
                # Case 4; fake/mmapped chunk
                self.arena = None
                self._memory_region = heap_region
                self._gdbValue = None

        self.start = self._memory_region.start
        # i686 alignment heuristic
        if Chunk(self.start).size == 0:
            self.start += pwndbg.aglib.arch.ptrsize * 2
        self.end = self._memory_region.end
        self.first_chunk = Chunk(self.start)

        self._prev = None

    @property
    def prev(self):
        if self._prev is None and self._gdbValue is not None:
            try:
                self._prev = int(self._gdbValue["prev"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._prev

    def __iter__(self):
        iter_chunk = self.first_chunk
        while iter_chunk is not None:
            yield iter_chunk
            iter_chunk = iter_chunk.next_chunk()

    def __contains__(self, addr: int) -> bool:
        return self.start <= addr < self.end

    def __str__(self) -> str:
        width = pwndbg.aglib.arch.ptrsize * 2
        return message.hint(f"[{hex(self.first_chunk.address):>{width}s}]") + mem_color.c.heap(
            str(pwndbg.aglib.vmmap.find(self.start))
        )


class Arena:
    __slots__ = (
        "_gdbValue",
        "address",
        "_is_main_arena",
        "_top",
        "_active_heap",
        "_heaps",
        "_mutex",
        "_flags",
        "_non_contiguous",
        "_have_fastchunks",
        "_fastbinsY",
        "_bins",
        "_binmap",
        "_next",
        "_next_free",
        "_system_mem",
    )

    def __init__(self, addr: int) -> None:
        assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
        assert pwndbg.aglib.heap.current.malloc_state is not None
        if isinstance(pwndbg.aglib.heap.current.malloc_state, pwndbg.dbg_mod.Type):
            self._gdbValue = pwndbg.aglib.memory.get_typed_pointer_value(
                pwndbg.aglib.heap.current.malloc_state, addr
            )
        else:
            self._gdbValue = pwndbg.aglib.heap.current.malloc_state(addr)

        self.address = int(self._gdbValue.address)
        self._is_main_arena: bool | None = None
        self._top = None
        self._active_heap = None
        self._heaps = None
        self._mutex = None
        self._flags = None
        self._non_contiguous = None
        self._have_fastchunks = None
        self._fastbinsY: list[int] | None = None
        self._bins: list[int] | None = None
        self._binmap: list[int] | None = None
        self._next: int | None = None
        self._next_free: int | None = None
        self._system_mem = None

    @property
    def is_main_arena(self) -> bool:
        assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
        if self._is_main_arena is None:
            self._is_main_arena = (
                pwndbg.aglib.heap.current.main_arena is not None
                and self.address == pwndbg.aglib.heap.current.main_arena.address
            )

        return self._is_main_arena

    @property
    def mutex(self) -> int | None:
        if self._mutex is None:
            try:
                self._mutex = int(self._gdbValue["mutex"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._mutex

    @property
    def flags(self) -> int | None:
        if self._flags is None:
            try:
                self._flags = int(self._gdbValue["flags"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._flags

    @property
    def non_contiguous(self) -> bool | None:
        if self._non_contiguous is None:
            flags = self.flags
            if flags is not None:
                self._non_contiguous = bool(flags & NONCONTIGUOUS_BIT)

        return self._non_contiguous

    @property
    def have_fastchunks(self) -> int | None:
        if self._have_fastchunks is None:
            try:
                self._have_fastchunks = int(self._gdbValue["have_fastchunks"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._have_fastchunks

    @property
    def top(self) -> int | None:
        if self._top is None:
            try:
                self._top = int(self._gdbValue["top"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._top

    @property
    def fastbinsY(self) -> list[int]:
        if self._fastbinsY is None:
            self._fastbinsY = []
            try:
                for i in range(NFASTBINS):
                    self._fastbinsY.append(int(self._gdbValue["fastbinsY"][i]))
            except pwndbg.dbg_mod.Error:
                pass

        return self._fastbinsY

    @property
    def bins(self) -> list[int]:
        if self._bins is None:
            self._bins = []
            try:
                for i in range(NBINS):
                    self._bins.append(int(self._gdbValue["bins"][i]))
            except pwndbg.dbg_mod.Error:
                pass

        return self._bins

    @property
    def binmap(self) -> list[int]:
        if self._binmap is None:
            self._binmap = []
            try:
                for i in range(BINMAPSIZE):
                    self._binmap.append(int(self._gdbValue["binmap"][i]))
            except pwndbg.dbg_mod.Error:
                pass

        return self._binmap

    @property
    def next(self) -> int | None:
        if self._next is None:
            try:
                self._next = int(self._gdbValue["next"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._next

    @property
    def next_free(self) -> int | None:
        if self._next_free is None:
            try:
                self._next_free = int(self._gdbValue["next_free"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._next_free

    @property
    def system_mem(self) -> int | None:
        if self._system_mem is None:
            try:
                self._system_mem = int(self._gdbValue["system_mem"])
            except pwndbg.dbg_mod.Error:
                pass

        return self._system_mem

    @property
    def active_heap(self) -> Heap:
        if self._active_heap is None:
            self._active_heap = Heap(self.top, arena=self)

        return self._active_heap

    @property
    def heaps(self):
        if self._heaps is None:
            heap = self.active_heap
            heap_list = [heap]
            if self.is_main_arena:
                assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
                sbrk_region = pwndbg.aglib.heap.current.get_sbrk_heap_region()
                if self.top not in sbrk_region:
                    heap_list.append(Heap(sbrk_region.start, arena=self))
            else:
                while heap.prev:
                    heap = Heap(heap.prev, arena=self)
                    heap_list.append(heap)

            heap_list.reverse()
            self._heaps = heap_list

        return self._heaps

    def fastbins(self) -> Bins:
        size = pwndbg.aglib.arch.ptrsize * 2
        fd_offset = pwndbg.aglib.arch.ptrsize * 2
        safe_lnk = pwndbg.libc.glibc.check_safe_linking(pwndbg.libc.version())
        result = Bins(BinType.FAST)
        for i in range(NFASTBINS):
            size += pwndbg.aglib.arch.ptrsize * 2
            chain = pwndbg.chain.get(
                int(self.fastbinsY[i]),
                offset=fd_offset,
                limit=pwndbg.aglib.heap.heap_chain_limit,
                safe_linking=safe_lnk,
            )

            result.bins[size] = Bin(chain)
        return result

    def __str__(self) -> str:
        width = pwndbg.aglib.arch.ptrsize * 2
        prefix_fmt = f"[{{:>{width}s}}]    "
        prefix_len = len(prefix_fmt.format(""))
        res = [message.hint(prefix_fmt.format(hex(self.address))) + str(self.heaps[0])]
        for h in self.heaps[1:]:
            res.append(" " * prefix_len + str(h))
        return "\n".join(res)


class GlibcMemoryAllocator(pwndbg.aglib.heap.heap.MemoryAllocator, Generic[TheType, TheValue]):
    # Largebin reverse lookup tables.
    # These help determine the range of chunk sizes that each largebin can hold.
    # They were generated by running every chunk size between the minimum & maximum large chunk
    # sizes through largebin_index().
    # Largebin 31 (bin 95) isn't used on i386 when MALLOC_ALIGNMENT is 16, so its value must be added manually.
    largebin_reverse_lookup_32 = (
        0x200,
        0x240,
        0x280,
        0x2C0,
        0x300,
        0x340,
        0x380,
        0x3C0,
        0x400,
        0x440,
        0x480,
        0x4C0,
        0x500,
        0x540,
        0x580,
        0x5C0,
        0x600,
        0x640,
        0x680,
        0x6C0,
        0x700,
        0x740,
        0x780,
        0x7C0,
        0x800,
        0x840,
        0x880,
        0x8C0,
        0x900,
        0x940,
        0x980,
        0x9C0,
        0xA00,
        0xC00,
        0xE00,
        0x1000,
        0x1200,
        0x1400,
        0x1600,
        0x1800,
        0x1A00,
        0x1C00,
        0x1E00,
        0x2000,
        0x2200,
        0x2400,
        0x2600,
        0x2800,
        0x2A00,
        0x3000,
        0x4000,
        0x5000,
        0x6000,
        0x7000,
        0x8000,
        0x9000,
        0xA000,
        0x10000,
        0x18000,
        0x20000,
        0x28000,
        0x40000,
        0x80000,
    )

    largebin_reverse_lookup_32_big = (
        0x3F0,
        0x400,
        0x440,
        0x480,
        0x4C0,
        0x500,
        0x540,
        0x580,
        0x5C0,
        0x600,
        0x640,
        0x680,
        0x6C0,
        0x700,
        0x740,
        0x780,
        0x7C0,
        0x800,
        0x840,
        0x880,
        0x8C0,
        0x900,
        0x940,
        0x980,
        0x9C0,
        0xA00,
        0xA40,
        0xA80,
        0xAC0,
        0xB00,
        0xB40,
        0xB80,  # Largebin 31 (bin 95) is unused, but its size is used to calculate the previous bin's maximum chunk size.
        0xB80,
        0xC00,
        0xE00,
        0x1000,
        0x1200,
        0x1400,
        0x1600,
        0x1800,
        0x1A00,
        0x1C00,
        0x1E00,
        0x2000,
        0x2200,
        0x2400,
        0x2600,
        0x2800,
        0x2A00,
        0x3000,
        0x4000,
        0x5000,
        0x6000,
        0x7000,
        0x8000,
        0x9000,
        0xA000,
        0x10000,
        0x18000,
        0x20000,
        0x28000,
        0x40000,
        0x80000,
    )

    largebin_reverse_lookup_64 = (
        0x400,
        0x440,
        0x480,
        0x4C0,
        0x500,
        0x540,
        0x580,
        0x5C0,
        0x600,
        0x640,
        0x680,
        0x6C0,
        0x700,
        0x740,
        0x780,
        0x7C0,
        0x800,
        0x840,
        0x880,
        0x8C0,
        0x900,
        0x940,
        0x980,
        0x9C0,
        0xA00,
        0xA40,
        0xA80,
        0xAC0,
        0xB00,
        0xB40,
        0xB80,
        0xBC0,
        0xC00,
        0xC40,
        0xE00,
        0x1000,
        0x1200,
        0x1400,
        0x1600,
        0x1800,
        0x1A00,
        0x1C00,
        0x1E00,
        0x2000,
        0x2200,
        0x2400,
        0x2600,
        0x2800,
        0x2A00,
        0x3000,
        0x4000,
        0x5000,
        0x6000,
        0x7000,
        0x8000,
        0x9000,
        0xA000,
        0x10000,
        0x18000,
        0x20000,
        0x28000,
        0x40000,
        0x80000,
    )

    def __init__(self) -> None:
        # Global ptmalloc objects
        self._global_max_fast_addr: int | None = None
        self._global_max_fast: int | None = None
        self._main_arena_addr: int | None = None
        self._main_arena: Arena | None = None
        self._mp_addr: int | None = None
        self._mp = None
        # List of arenas/heaps
        self._arenas = None
        # ptmalloc cache for current thread
        self._thread_cache: TheValue | None = None

    def largebin_reverse_lookup(self, index: int) -> int:
        """Pick the appropriate largebin_reverse_lookup_ function for this architecture."""
        if pwndbg.aglib.arch.ptrsize == 8:
            return self.largebin_reverse_lookup_64[index]
        if self.malloc_alignment == 16:
            return self.largebin_reverse_lookup_32_big[index]
        return self.largebin_reverse_lookup_32[index]

    def largebin_size_range_from_index(self, index: int):
        largest_largebin = self.largebin_index(pwndbg.aglib.arch.ptrmask) - 64
        start_size = self.largebin_reverse_lookup(index)

        if index != largest_largebin:
            end_size = self.largebin_reverse_lookup(index + 1) - self.malloc_alignment
        else:
            end_size = pwndbg.aglib.arch.ptrmask

        return (start_size, end_size)

    def can_be_resolved(self) -> bool:
        raise NotImplementedError()

    @property
    def main_arena(self) -> Arena | None:
        raise NotImplementedError()

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def arenas(self) -> tuple[Arena, ...]:
        """Return a tuple of all current arenas."""
        arenas: list[Arena] = []
        main_arena = self.main_arena
        if main_arena:
            arenas.append(main_arena)

            arena = main_arena
            addr = arena.next
            while addr is not None and addr != main_arena.address:
                arena = Arena(addr)
                arenas.append(arena)
                addr = arena.next

        self._arenas = tuple(arenas)
        return self._arenas

    def has_tcache(self) -> bool:
        raise NotImplementedError()

    @property
    def thread_arena(self) -> Arena | None:
        raise NotImplementedError()

    @property
    def thread_cache(self) -> TheValue | None:
        raise NotImplementedError()

    @property
    def mp(self) -> TheValue | None:
        raise NotImplementedError()

    @property
    def global_max_fast(self) -> int | None:
        raise NotImplementedError()

    @property
    def heap_info(self) -> TheType | None:
        raise NotImplementedError()

    @property
    def malloc_chunk(self) -> TheType | None:
        raise NotImplementedError()

    @property
    def malloc_state(self) -> TheType | None:
        raise NotImplementedError()

    @property
    def tcache_perthread_struct(self) -> TheType | None:
        raise NotImplementedError()

    @property
    def tcache_entry(self) -> TheType | None:
        raise NotImplementedError()

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def tcache_small_bins(self) -> int | None:
        if not self.has_tcache():
            return None
        mp = self.mp
        if "tcache_small_bins" in mp.type.keys():
            return int(mp["tcache_small_bins"])
        if "tcache_bins" in mp.type.keys():
            return int(mp["tcache_bins"])
        return None

    @property
    def mallinfo(self) -> TheType | None:
        raise NotImplementedError()

    @property
    def malloc_par(self) -> TheType | None:
        raise NotImplementedError()

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_alignment(self) -> int:
        """Corresponds to MALLOC_ALIGNMENT in glibc malloc.c"""
        if pwndbg.aglib.arch.name == "i386" and pwndbg.libc.version() >= (2, 26):
            # i386 will override it to 16 when GLIBC version >= 2.26
            # See https://elixir.bootlin.com/glibc/glibc-2.26/source/sysdeps/i386/malloc-alignment.h#L22
            return 16
        # See https://elixir.bootlin.com/glibc/glibc-2.37/source/sysdeps/generic/malloc-alignment.h#L27
        long_double_alignment = pwndbg.aglib.typeinfo.lookup_types("long double").alignof
        return (
            long_double_alignment if 2 * self.size_sz < long_double_alignment else 2 * self.size_sz
        )

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def size_sz(self) -> int:
        """Corresponds to SIZE_SZ in glibc malloc.c"""
        return pwndbg.aglib.arch.ptrsize

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_align_mask(self) -> int:
        """Corresponds to MALLOC_ALIGN_MASK in glibc malloc.c"""
        return self.malloc_alignment - 1

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def minsize(self) -> int:
        """Corresponds to MINSIZE in glibc malloc.c"""
        return self.min_chunk_size

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def min_chunk_size(self) -> int:
        """Corresponds to MIN_CHUNK_SIZE in glibc malloc.c"""
        return pwndbg.aglib.arch.ptrsize * 4

    @property
    @pwndbg.lib.cache.cache_until("objfile", "thread")
    def multithreaded(self) -> bool:
        """Is malloc operating within a multithreaded environment."""
        addr = pwndbg.aglib.symbol.lookup_symbol_addr("__libc_multiple_threads")
        if addr:
            return pwndbg.aglib.memory.u32(addr) > 0
        # glibc 2.42 replaced __libc_multiple_threads with __libc_single_threaded
        if addr := pwndbg.aglib.symbol.lookup_symbol_addr("__libc_single_threaded"):
            return pwndbg.aglib.memory.u32(addr) == 0
        return len(pwndbg.dbg.selected_inferior().threads()) > 1

    def _request2size(self, req: int) -> int:
        """Corresponds to request2size in glibc malloc.c"""
        if req + self.size_sz + self.malloc_align_mask < self.minsize:
            return self.minsize
        return (req + self.size_sz + self.malloc_align_mask) & ~self.malloc_align_mask

    def chunk_flags(self, size: int) -> tuple[int, int, int]:
        return (
            size & PREV_INUSE,
            size & IS_MMAPPED,
            size & NON_MAIN_ARENA,
        )

    def chunk_key_offset(self, key: str) -> int | None:
        """Find the index of a field in the malloc_chunk struct.

        64bit example:
            prev_size == 0
            size      == 8
            fd        == 16
            bk        == 24
            ...
        """
        renames = {
            "mchunk_size": "size",
            "mchunk_prev_size": "prev_size",
        }
        val = self.malloc_chunk
        if val is None:
            return None
        chunk_keys = [renames[key] if key in renames else key for key in val.keys()]
        try:
            return chunk_keys.index(key) * pwndbg.aglib.arch.ptrsize
        except Exception:
            return None

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def tcache_next_offset(self) -> int:
        return self.tcache_entry.keys().index("next") * pwndbg.aglib.arch.ptrsize

    def get_heap(self, addr: int) -> TheValue | None:
        raise NotImplementedError()

    def get_tcache(self, tcache_addr: int | None = None) -> TheValue | None:
        raise NotImplementedError()

    def get_sbrk_heap_region(self) -> pwndbg.lib.memory.Page | None:
        raise NotImplementedError()

    def get_region(self, addr: int | pwndbg.dbg_mod.Value | None) -> pwndbg.lib.memory.Page | None:
        """Find the memory map containing 'addr'."""
        return copy.deepcopy(pwndbg.aglib.vmmap.find(addr))

    def get_bins(self, bin_type: BinType, addr: int | None = None) -> Bins | None:
        if bin_type == BinType.TCACHE:
            return self.tcachebins(addr)
        if bin_type == BinType.FAST:
            return self.fastbins(addr)
        if bin_type == BinType.UNSORTED:
            return self.unsortedbin(addr)
        if bin_type == BinType.SMALL:
            return self.smallbins(addr)
        if bin_type == BinType.LARGE:
            return self.largebins(addr)
        return None

    def fastbin_index(self, size: int):
        if pwndbg.aglib.arch.ptrsize == 8:
            return (size >> 4) - 2
        return (size >> 3) - 2

    def fastbins(self, arena_addr: int | None = None) -> Bins | None:
        """Returns: chain or None"""
        if arena_addr:
            arena = Arena(arena_addr)
        else:
            arena = self.thread_arena

        if arena is None:
            return None

        return arena.fastbins()

    def tcachebins(self, tcache_addr: int | None = None) -> Bins | None:
        """Returns: tuple(chain, count) or None"""
        tcache = self.get_tcache(tcache_addr)

        if tcache is None:
            return None

        # this will break expected output during tests, so we skip it
        if (
            pwndbg.libc.version() >= (2, 42)
            and not hasattr(GlibcMemoryAllocator.tcachebins, "tcache_2_42_warning_issued")
            and os.environ.get("PWNDBG_IN_TEST") is None
        ):
            print(
                message.warn(
                    "Support for tcache large bins (a GLIBC 2.42 addition) has not been fully implemented. "
                    "PR contributions are highly appreciated!"
                )
            )
            setattr(GlibcMemoryAllocator.tcachebins, "tcache_2_42_warning_issued", True)

        # counts was renamed to num_slots in newer version of GLIBC 2.42
        try:
            counts = tcache["num_slots"]
        except Exception:
            counts = tcache["counts"]
        entries = tcache["entries"]

        num_tcachebins = entries.type.sizeof // entries.type.target().sizeof
        safe_lnk = pwndbg.libc.glibc.check_safe_linking(pwndbg.libc.version())

        def tidx2usize(idx: int):
            """Tcache bin index to chunk size, following tidx2usize macro in glibc malloc.c"""
            return idx * self.malloc_alignment + self.minsize - self.size_sz

        result = Bins(BinType.TCACHE)
        for i in range(num_tcachebins):
            size = self._request2size(tidx2usize(i))
            count = int(counts[i])
            if pwndbg.libc.version() >= (2, 42):
                count = int(self.mp["tcache_count"]) - count
            chain = pwndbg.chain.get(
                int(entries[i]),
                offset=self.tcache_next_offset,
                limit=pwndbg.aglib.heap.heap_chain_limit,
                safe_linking=safe_lnk,
            )

            result.bins[size] = Bin(chain, count=count)
        return result

    def check_chain_corrupted(self, chain_fd: list[int], chain_bk: list[int]) -> bool:
        """
        Checks if the doubly linked list (of a {unsorted, small, large} bin)
        defined by chain_fd, chain_bk is corrupted.

        Even if the chains do not cover the whole bin, they still are expected
        to be of the same length.

        Returns True if the bin is certainly corrupted, otherwise False.
        """

        if len(chain_fd) != len(chain_bk):
            # If the chain lengths aren't equal, the chain is corrupted
            # The vast majority of corruptions will be caught here
            return True
        if len(chain_fd) < 2 or len(chain_bk) < 2:
            # Chains containing less than two entries are corrupted, as the smallest
            # chain (an empty bin) would look something like `[main_arena+88, 0]`.
            return True
        if len(chain_fd) == len(chain_bk) == 2:
            # Check if the bin points to itself (is empty)

            if chain_fd != chain_bk:
                return True
            if chain_fd[-1] != 0:
                return True
            bin_chk = Chunk(chain_fd[0])
            if not (bin_chk.fd == bin_chk.bk == chain_fd[0]):
                return True

        else:
            chain_sz = len(chain_fd) - (1 if chain_fd[-1] == 0 else 0)

            # Forward and backward chains may have some overlap, we don't need to recheck those chunks
            checked = set()

            # Check connections in all chunks from the forward chain
            for i in range(chain_sz):
                chunk_addr = chain_fd[i]
                chunk = Chunk(chunk_addr)
                if chunk.fd is None or Chunk(chunk.fd).bk != chunk_addr:
                    return True
                if chunk.bk is None or Chunk(chunk.bk).fd != chunk_addr:
                    return True
                checked.add(chunk_addr)

            # Check connections in unchecked chunks from the backward chain
            for i in range(chain_sz):
                chunk_addr = chain_bk[i]
                if chunk_addr in checked:
                    # We don't need to check any more chunks
                    break
                chunk = Chunk(chunk_addr)
                if chunk.fd is None or Chunk(chunk.fd).bk != chunk_addr:
                    return True
                if chunk.bk is None or Chunk(chunk.bk).fd != chunk_addr:
                    return True

        return False

    def bin_at(
        self, index: int, arena_addr: int | None = None
    ) -> tuple[list[int], list[int], bool] | None:
        """
        Modeled after glibc's bin_at function - so starts indexing from 1
        https://bazaar.launchpad.net/~ubuntu-branches/ubuntu/trusty/eglibc/trusty-security/view/head:/malloc/malloc.c#L1394

        bin_at(1) returns the unsorted bin

        Bin 1          - Unsorted BiN
        Bin 2 to 63    - Smallbins
        Bin 64 to 126  - Largebins

        Returns: tuple(chain_from_bin_fd, chain_from_bin_bk, is_chain_corrupted) or None
        """
        index = index - 1

        if arena_addr is not None:
            arena = Arena(arena_addr)
        else:
            arena = self.thread_arena

        if arena is None:
            return None

        normal_bins = arena._gdbValue["bins"]  # Breaks encapsulation, find a better way.

        bins_base = int(normal_bins.address) - (pwndbg.aglib.arch.ptrsize * 2)
        current_base = bins_base + (index * pwndbg.aglib.arch.ptrsize * 2)

        # check whether the bin is empty
        bin_chunk = Chunk(current_base)
        if bin_chunk.fd == bin_chunk.bk == current_base:
            return ([0], [0], False)

        front, back = normal_bins[index * 2], normal_bins[index * 2 + 1]
        fd_offset = self.chunk_key_offset("fd")
        bk_offset = self.chunk_key_offset("bk")

        chain_size = int(pwndbg.aglib.heap.heap_chain_limit)
        corrupt_chain_size = int(pwndbg.aglib.heap.heap_corruption_check_limit)

        get_chain = lambda bin, offset: pwndbg.chain.get(
            int(bin),
            offset=offset,
            hard_stop=current_base,
            limit=max(chain_size, corrupt_chain_size),
            include_start=True,
        )

        full_chain_fd = get_chain(front, fd_offset)
        full_chain_bk = get_chain(back, bk_offset)
        chain_fd = full_chain_fd[: (chain_size + 1)]
        chain_bk = full_chain_bk[: (chain_size + 1)]
        corrupt_chain_fd = full_chain_fd[: (corrupt_chain_size + 1)]
        corrupt_chain_bk = full_chain_bk[: (corrupt_chain_size + 1)]

        is_chain_corrupted = False
        if corrupt_chain_size > 1:
            is_chain_corrupted = self.check_chain_corrupted(corrupt_chain_fd, corrupt_chain_bk)

        return (chain_fd, chain_bk, is_chain_corrupted)

    def unsortedbin(self, arena_addr: int | None = None) -> Bins | None:
        chain = self.bin_at(1, arena_addr=arena_addr)
        result = Bins(BinType.UNSORTED)

        if chain is None:
            return None

        fd_chain, bk_chain, is_corrupted = chain
        result.bins["all"] = Bin(fd_chain, bk_chain, is_corrupted=is_corrupted)
        return result

    def smallbins(self, arena_addr: int | None = None) -> Bins | None:
        size = self.min_chunk_size
        result = Bins(BinType.SMALL)
        for index in range(2, 64):
            chain = self.bin_at(index, arena_addr=arena_addr)

            if chain is None:
                return None

            fd_chain, bk_chain, is_corrupted = chain
            result.bins[size] = Bin(fd_chain, bk_chain, is_corrupted=is_corrupted)
            size += self.malloc_alignment
        return result

    def largebins(self, arena_addr: int | None = None) -> Bins | None:
        result = Bins(BinType.LARGE)
        for index in range(64, 127):
            chain = self.bin_at(index, arena_addr=arena_addr)

            if chain is None:
                return None

            fd_chain, bk_chain, is_corrupted = chain
            result.bins[index - NSMALLBINS] = Bin(fd_chain, bk_chain, is_corrupted=is_corrupted)

        return result

    def largebin_index_32(self, sz: int) -> int:
        """Modeled on the GLIBC malloc largebin_index_32 macro.

        https://sourceware.org/git/?p=glibc.git;a=blob;f=malloc/malloc.c;h=f7cd29bc2f93e1082ee77800bd64a4b2a2897055;hb=9ea3686266dca3f004ba874745a4087a89682617#l1414
        """
        return (
            56 + (sz >> 6)
            if (sz >> 6) <= 38
            else (
                91 + (sz >> 9)
                if (sz >> 9) <= 20
                else (
                    110 + (sz >> 12)
                    if (sz >> 12) <= 10
                    else (
                        119 + (sz >> 15)
                        if (sz >> 15) <= 4
                        else 124 + (sz >> 18)
                        if (sz >> 18) <= 2
                        else 126
                    )
                )
            )
        )

    def largebin_index_32_big(self, sz: int) -> int:
        """Modeled on the GLIBC malloc largebin_index_32_big macro.

        https://sourceware.org/git/?p=glibc.git;a=blob;f=malloc/malloc.c;h=f7cd29bc2f93e1082ee77800bd64a4b2a2897055;hb=9ea3686266dca3f004ba874745a4087a89682617#l1422
        """
        return (
            49 + (sz >> 6)
            if (sz >> 6) <= 45
            else (
                91 + (sz >> 9)
                if (sz >> 9) <= 20
                else (
                    110 + (sz >> 12)
                    if (sz >> 12) <= 10
                    else (
                        119 + (sz >> 15)
                        if (sz >> 15) <= 4
                        else 124 + (sz >> 18)
                        if (sz >> 18) <= 2
                        else 126
                    )
                )
            )
        )

    def largebin_index_64(self, sz: int) -> int:
        """Modeled on the GLIBC malloc largebin_index_64 macro.

        https://sourceware.org/git/?p=glibc.git;a=blob;f=malloc/malloc.c;h=f7cd29bc2f93e1082ee77800bd64a4b2a2897055;hb=9ea3686266dca3f004ba874745a4087a89682617#l1433
        """
        return (
            48 + (sz >> 6)
            if (sz >> 6) <= 48
            else (
                91 + (sz >> 9)
                if (sz >> 9) <= 20
                else (
                    110 + (sz >> 12)
                    if (sz >> 12) <= 10
                    else (
                        119 + (sz >> 15)
                        if (sz >> 15) <= 4
                        else 124 + (sz >> 18)
                        if (sz >> 18) <= 2
                        else 126
                    )
                )
            )
        )

    def largebin_index(self, sz: int):
        """Pick the appropriate largebin_index_ function for this architecture."""
        if pwndbg.aglib.arch.ptrsize == 8:
            return self.largebin_index_64(sz)
        if self.malloc_alignment == 16:
            return self.largebin_index_32_big(sz)
        return self.largebin_index_32(sz)

    def is_initialized(self):
        raise NotImplementedError()

    def is_statically_linked(self) -> bool:
        return not pwndbg.dbg.selected_inferior().is_dynamically_linked()


class DebugSymsHeap(GlibcMemoryAllocator[pwndbg.dbg_mod.Type, pwndbg.dbg_mod.Value]):
    def can_be_resolved(self) -> bool:
        if not pwndbg.libc.has_debug_info():
            return False
        # Check if thread_arena is needed and available, but if the binary is not multithreaded, then we don't care
        # Note: it's possible that we unstripped the libc but still don't have libthread_db.so
        return (
            not self.multithreaded
            or pwndbg.aglib.symbol.lookup_symbol_addr("thread_arena", prefer_static=True)
            is not None
        )

    @property
    def main_arena(self) -> Arena | None:
        self._main_arena_addr = pwndbg.aglib.symbol.lookup_symbol_addr(
            "main_arena", prefer_static=True
        )
        if self._main_arena_addr is not None:
            self._main_arena = Arena(self._main_arena_addr)

        return self._main_arena

    def has_tcache(self) -> bool:
        # tcache_bins was renamed to tcache_small_bins in GLIBC 2.42
        return self.mp is not None and any(
            x in self.mp.type.keys() for x in ["tcache_bins", "tcache_small_bins"]
        )

    @property
    def thread_arena(self) -> Arena | None:
        if self.multithreaded:
            thread_arena_addr = pwndbg.aglib.symbol.lookup_symbol_addr(
                "thread_arena", prefer_static=True
            )
            if thread_arena_addr:
                thread_arena_value = pwndbg.aglib.memory.read_pointer_width(thread_arena_addr)
                # thread_arena might be NULL if the thread doesn't allocate arena yet
                if thread_arena_value:
                    return Arena(pwndbg.aglib.memory.read_pointer_width(thread_arena_addr))
            return None
        return self.main_arena

    @property
    def thread_cache(self) -> pwndbg.dbg_mod.Value | None:
        """Locate a thread's tcache struct. If it doesn't have one, use the main
        thread's tcache.
        """
        if not self.has_tcache():
            print(message.warn("This version of GLIBC was not compiled with tcache support."))
            return None

        tcache_ptr = pwndbg.aglib.symbol.lookup_symbol_addr(
            "tcache",
            prefer_static=True,
        )
        if tcache_ptr and (tcache_addr := pwndbg.aglib.memory.read_pointer_width(tcache_ptr)):
            tcache = tcache_addr
        elif not self.multithreaded:
            tcache = self.main_arena.heaps[0].start + pwndbg.aglib.arch.ptrsize * 2
        else:
            # This thread doesn't have a tcache yet
            return None

        try:
            self._thread_cache = pwndbg.aglib.memory.get_typed_pointer_value(
                self.tcache_perthread_struct, tcache
            )
            self._thread_cache["entries"].fetch_lazy()
        except Exception:
            print(
                message.error(
                    "Error fetching tcache. Cannot access "
                    "thread-local variables unless you compile with -lpthread."
                )
            )
            return None

        return self._thread_cache

    @property
    def mp(self) -> pwndbg.dbg_mod.Value | None:
        self._mp_addr = pwndbg.aglib.symbol.lookup_symbol_addr("mp_", prefer_static=True)
        if self._mp_addr is not None and self.malloc_par is not None:
            self._mp = pwndbg.aglib.memory.get_typed_pointer_value(self.malloc_par, self._mp_addr)

        return self._mp

    @property
    def global_max_fast(self) -> int | None:
        self._global_max_fast_addr = pwndbg.aglib.symbol.lookup_symbol_addr(
            "global_max_fast", prefer_static=True
        )
        if self._global_max_fast_addr is not None:
            self._global_max_fast = pwndbg.aglib.memory.u(self._global_max_fast_addr)

        return self._global_max_fast

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def heap_info(self) -> pwndbg.dbg_mod.Type | None:
        return pwndbg.aglib.typeinfo.load("heap_info")

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_chunk(self) -> pwndbg.dbg_mod.Type | None:
        return pwndbg.aglib.typeinfo.load("struct malloc_chunk")

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_state(self) -> pwndbg.dbg_mod.Type | None:
        return pwndbg.aglib.typeinfo.load("struct malloc_state")

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def tcache_perthread_struct(self) -> pwndbg.dbg_mod.Type | None:
        return pwndbg.aglib.typeinfo.load("struct tcache_perthread_struct")

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def tcache_entry(self) -> pwndbg.dbg_mod.Type | None:
        return pwndbg.aglib.typeinfo.load("struct tcache_entry")

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def mallinfo(self) -> pwndbg.dbg_mod.Type | None:
        return pwndbg.aglib.typeinfo.load("struct mallinfo")

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_par(self) -> pwndbg.dbg_mod.Type | None:
        return pwndbg.aglib.typeinfo.load("struct malloc_par")

    def get_heap(self, addr: int) -> pwndbg.dbg_mod.Value | None:
        """Find & read the heap_info struct belonging to the chunk at 'addr'."""
        if self.heap_info is None:
            return None
        haddr = heap_for_ptr(addr)
        if pwndbg.aglib.memory.peek(haddr) is None:
            return None
        return pwndbg.aglib.memory.get_typed_pointer_value(self.heap_info, haddr)

    def get_tcache(
        self, tcache_addr: int | pwndbg.dbg_mod.Value | None = None
    ) -> pwndbg.dbg_mod.Value | None:
        if tcache_addr is None:
            return self.thread_cache

        return pwndbg.aglib.memory.get_typed_pointer_value(
            self.tcache_perthread_struct, tcache_addr
        )

    def get_sbrk_heap_region(self) -> pwndbg.lib.memory.Page | None:
        """Return a Page object representing the sbrk heap region.
        Ensure the region's start address is aligned to SIZE_SZ * 2,
        which compensates for the presence of GLIBC_TUNABLES.
        """
        assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
        assert self.mp is not None
        sbrk_base = pwndbg.lib.memory.align_up(
            int(self.mp["sbrk_base"]), pwndbg.aglib.heap.current.size_sz * 2
        )

        sbrk_region = self.get_region(sbrk_base)
        if sbrk_region is None:
            return None
        sbrk_region.memsz = sbrk_region.end - sbrk_base
        sbrk_region.vaddr = sbrk_base

        return sbrk_region

    def is_initialized(self) -> bool:
        addr = pwndbg.aglib.symbol.lookup_symbol_addr("__libc_malloc_initialized")
        if addr is None:
            addr = pwndbg.aglib.symbol.lookup_symbol_addr("__malloc_initialized")
        # fallback for GLIBC 2.42 as __malloc_initialized was removed
        if addr is None:
            return int(self.mp["sbrk_base"]) != 0
        return pwndbg.aglib.memory.s32(addr) > 0


class HeuristicHeap(
    GlibcMemoryAllocator[
        type["pwndbg.aglib.heap.structs.CStruct2GDB"],
        "pwndbg.aglib.heap.structs.CStruct2GDB",
    ]
):
    def __init__(self) -> None:
        super().__init__()
        self._structs_module: types.ModuleType | None = None
        self._thread_arena_values: dict[int, int] = {}
        self._thread_caches: dict[int, Any] = {}

    @property
    def struct_module(self) -> types.ModuleType | None:
        if not self._structs_module and pwndbg.libc.version() != (-1, -1):
            try:
                self._structs_module = importlib.reload(
                    importlib.import_module("pwndbg.aglib.heap.structs")
                )
            except AssertionError:
                raise
            except Exception:
                pass
        return self._structs_module

    def can_be_resolved(self) -> bool:
        return self.struct_module is not None

    @property
    def main_arena(self) -> Arena | None:
        main_arena_via_config = int(str(pwndbg.config.main_arena), 0)
        main_arena_via_symbol = pwndbg.aglib.symbol.lookup_symbol_addr(
            "main_arena", prefer_static=True
        )
        if main_arena_via_config or main_arena_via_symbol:
            self._main_arena_addr = main_arena_via_config or main_arena_via_symbol

        if not self._main_arena_addr:
            if self.is_statically_linked():
                data_section = pwndbg.aglib.proc.dump_elf_data_section()
                data_section_address = pwndbg.aglib.proc.get_section_address_by_name(".data")
            else:
                data_section = pwndbg.libc.section_by_name(".data")
                data_section_address = pwndbg.libc.section_address_by_name(".data")
            if data_section and data_section_address:
                data_section_offset, size, data_section_data = data_section
                # Try to find the default main_arena struct in the .data section
                # https://github.com/bminor/glibc/blob/glibc-2.37/malloc/malloc.c#L1902-L1907
                # static struct malloc_state main_arena =
                # {
                #   .mutex = _LIBC_LOCK_INITIALIZER,
                #   .next = &main_arena,
                #   .attached_threads = 1
                # };
                expected = self.malloc_state._c_struct()
                expected.attached_threads = 1
                next_field_offset = self.malloc_state.get_field_offset("next")
                malloc_state_size = self.malloc_state.sizeof

                # Since RELR relocations might also have .rela.dyn section, we check it first
                for section_name in (".relr.dyn", ".rela.dyn", ".rel.dyn"):
                    if self._main_arena_addr:
                        # If we have found the main_arena, we can stop searching
                        break

                    if self.is_statically_linked():
                        relocations = pwndbg.aglib.proc.dump_relocations_by_section_name(
                            section_name
                        )
                    else:
                        relocations = pwndbg.libc.relocations_by_section_name(section_name)
                    if not relocations:
                        continue

                    for relocation in relocations:
                        r_offset = relocation.entry.r_offset

                        # We only care about the relocation in .data section
                        if r_offset - next_field_offset < data_section_offset:
                            continue

                        if r_offset - next_field_offset >= data_section_offset + size:
                            break

                        # To find addend:
                        # .relr.dyn and .rel.dyn need to read the data from r_offset
                        # .rela.dyn has the addend in the entry
                        if section_name != ".rela.dyn":
                            addend = int.from_bytes(
                                data_section_data[
                                    r_offset - data_section_offset : r_offset
                                    - data_section_offset
                                    + pwndbg.aglib.arch.ptrsize
                                ],
                                pwndbg.aglib.arch.endian,
                            )
                        else:
                            addend = relocation.entry.r_addend

                        # If addend is the offset of main_arena, then r_offset should be the offset of main_arena.next
                        if r_offset - next_field_offset == addend:
                            # Check if we can construct the default main_arena struct we expect
                            tmp = data_section_data[
                                addend - data_section_offset : addend
                                - data_section_offset
                                + malloc_state_size
                            ]
                            # Note: Although RELA relocations have r_addend, some compiler will still put the addend in the location of r_offset, so we still need to check both cases
                            found = False
                            expected.next = addend
                            found |= bytes(expected) == tmp
                            if not found:
                                expected.next = 0
                                found |= bytes(expected) == tmp
                            if found:
                                # This might be a false positive, but it is very unlikely, so should be fine :)
                                self._main_arena_addr = (
                                    data_section_address + addend - data_section_offset
                                )
                                break

                # If we are still not able to find the main_arena, probably we are debugging a binary with statically linked libc and no PIE enabled
                if not self._main_arena_addr:
                    # Try to find the default main_arena struct in the .data section
                    for i in range(0, size - self.malloc_state.sizeof, pwndbg.aglib.arch.ptrsize):
                        expected.next = data_section_offset + i
                        if bytes(expected) == data_section_data[i : i + malloc_state_size]:
                            # This also might be a false positive, but it is very unlikely too, so should also be fine :)
                            self._main_arena_addr = data_section_address + i
                            break

        if pwndbg.aglib.memory.is_readable_address(self._main_arena_addr):
            self._main_arena = Arena(self._main_arena_addr)
            return self._main_arena

        raise SymbolNotRecoveredError("main_arena", "Heuristic failed.")

    def has_tcache(self) -> bool:
        # TODO/FIXME: Can we determine the tcache_bins existence more reliable?

        # There is no debug symbols, we determine the tcache_bins existence by checking glibc version only
        return self.is_initialized() and pwndbg.libc.version() >= (2, 26)

    def prompt_for_brute_force_thread_arena_permission(self) -> bool:
        """Check if the user wants to brute force the thread_arena's value."""
        print(
            message.notice("We cannot determine the {}\n".format(message.hint("thread_arena")))
            + message.notice(
                "Will you want to brute force it in the memory to determine the address? (y/N)\n"
            )
            + message.warn(
                "Note: This might take a while and might not be reliable, so if you can determine it by yourself or you have modified any of the arena, please do not use this."
            )
        )
        return input().lower() == "y"

    def prompt_for_brute_force_thread_cache_permission(self) -> bool:
        """Check if the user wants to brute force the tcache's value."""
        print(
            message.notice("We cannot determine the {}\n".format(message.hint("tcache")))
            + message.notice(
                "Will you want to brute force it in the memory to determine the address instead of assuming it's at the beginning of the current thread's heap? (y/N)\n"
            )
            + message.warn(
                "Note: This might take a while and might not be reliable, so if you can determine it by yourself or your current arena is corrupted or you have modified the chunk for the tcache, please do not use this."
            )
        )
        return input().lower() == "y"

    def prompt_for_tls_address(self) -> int:
        """Check if we can determine the TLS address and return it."""
        tls_address = pwndbg.aglib.tls.find_address_with_register()
        if not tls_address:
            print(
                message.warn("Cannot find TLS address via register. ")
                + message.notice(
                    "Will you want to call pthread_self() to find the address? (y/N)\n"
                )
                + message.warn("Note: Don't use this if pthread_self() is not available.")
            )
            if input().lower() == "y":
                tls_address = pwndbg.aglib.tls.find_address_with_pthread_self()
            if not tls_address:
                print(message.error("Cannot find TLS address via pthread_self()."))
        return tls_address

    def brute_force_tls_reference_in_got_section(
        self, tls_address: int, validator: Callable[[int], bool]
    ) -> tuple[int, int] | None:
        """Brute force the TLS-reference in the .got section to that can pass the validator."""
        # Note: This highly depends on the correctness of the TLS address
        print(message.notice("Brute forcing the TLS-reference in the .got section..."))
        if self.is_statically_linked():
            got_address = pwndbg.aglib.proc.get_section_address_by_name(".got")
        else:
            got_address = pwndbg.libc.section_address_by_name(".got")
        if not got_address:
            print(message.warn("Cannot find the address of the .got section."))
            return None
        s_int = (
            pwndbg.aglib.memory.s32 if pwndbg.aglib.arch.ptrsize == 4 else pwndbg.aglib.memory.s64
        )
        for addr in range(got_address, got_address + 0xF0, pwndbg.aglib.arch.ptrsize):
            if not pwndbg.aglib.memory.is_readable_address(addr):
                break
            offset = s_int(addr)
            if (
                offset
                and offset % pwndbg.aglib.arch.ptrsize == 0
                and pwndbg.aglib.memory.is_readable_address(offset + tls_address)
            ):
                guess = pwndbg.aglib.memory.read_pointer_width(offset + tls_address)
                if validator(guess):
                    return guess, offset + tls_address
        return None

    def brute_force_thread_local_variable_near_tls_base(
        self, tls_address: int, validator: Callable[[int], bool]
    ) -> tuple[int, int] | None:
        """Brute force the thread-local variable near the TLS base address that can pass the validator."""
        print(
            message.notice(
                "Brute forcing all the possible thread-local variables near the TLS base address..."
            )
        )
        for search_range in (
            range(tls_address, tls_address - 0x500, -pwndbg.aglib.arch.ptrsize),
            range(tls_address, tls_address + 0x500, pwndbg.aglib.arch.ptrsize),
        ):
            reading = False
            for addr in search_range:
                if pwndbg.aglib.memory.is_readable_address(addr):
                    reading = True
                    guess = pwndbg.aglib.memory.read_pointer_width(addr)
                    if validator(guess):
                        return guess, addr
                elif reading:
                    # Don't need to try now, we only read consecutive memory
                    break
        return None

    @property
    def thread_arena(self) -> Arena | None:
        thread_arena_via_symbol = pwndbg.aglib.symbol.lookup_symbol_addr(
            "thread_arena", prefer_static=True
        )
        if thread_arena_via_symbol:
            thread_arena_value = pwndbg.aglib.memory.read_pointer_width(thread_arena_via_symbol)
            return Arena(thread_arena_value) if thread_arena_value else None
        thread_arena_via_config = int(str(pwndbg.config.thread_arena), 0)
        if thread_arena_via_config:
            return Arena(thread_arena_via_config)

        # return the value of the thread_arena if we have it cached
        thread_arena_value = self._thread_arena_values.get(pwndbg.dbg.selected_thread().index())
        if thread_arena_value:
            return Arena(thread_arena_value)

        assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
        if (
            self.main_arena.address != pwndbg.aglib.heap.current.main_arena.next
            or self.multithreaded
        ):
            if pwndbg.aglib.arch.name not in ("i386", "x86-64", "arm", "aarch64"):
                # TODO: Support other architectures
                raise SymbolNotRecoveredError(
                    "thread_arena", "Unsupported architecture for thread arena bruteforce."
                )
            if self.prompt_for_brute_force_thread_arena_permission():
                tls_address = self.prompt_for_tls_address()
                if not tls_address:
                    raise SymbolNotRecoveredError("thread_arena", "TLS not found.")
                print(message.notice("Fetching all the arena addresses..."))
                candidates = [a.address for a in self.arenas]

                def validator(guess: int) -> bool:
                    return guess in candidates

                found = self.brute_force_tls_reference_in_got_section(
                    tls_address, validator
                ) or self.brute_force_thread_local_variable_near_tls_base(tls_address, validator)
                if found:
                    value, address = found
                    print(
                        message.notice(
                            f"Found matching arena address {message.hint(hex(value))} at {message.hint(hex(address))}\n"
                        )
                    )
                    arena = Arena(value)
                    self._thread_arena_values[pwndbg.dbg.selected_thread().index()] = value
                    return arena

                print(
                    message.notice(
                        f"Cannot find {message.hint('thread_arena')}, the arena might be not allocated yet.\n"
                    )
                )
                return None
            raise SymbolNotRecoveredError("thread_arena", "Thread arena bruteforce failed.")
        else:  # noqa: RET506
            self._thread_arena_values[pwndbg.dbg.selected_thread().index()] = (
                self.main_arena.address
            )
            return self.main_arena

    @property
    def thread_cache(self) -> pwndbg.aglib.heap.structs.TcachePerthreadStruct | None:
        """Locate a thread's tcache struct. We try to find its address in Thread Local Storage (TLS) first,
        and if that fails, we guess it's at the first chunk of the heap.
        """
        if not self.has_tcache():
            print(message.warn("This version of GLIBC was not compiled with tcache support."))
            return None
        tps = self.tcache_perthread_struct
        thread_cache_via_config = int(str(pwndbg.config.tcache), 0)
        thread_cache_via_symbol = pwndbg.aglib.symbol.lookup_symbol_addr(
            "tcache", prefer_static=True
        )
        if thread_cache_via_config:
            self._thread_cache = tps(thread_cache_via_config)
            return self._thread_cache
        if thread_cache_via_symbol:
            thread_cache_struct_addr = pwndbg.aglib.memory.read_pointer_width(
                thread_cache_via_symbol
            )
            if thread_cache_struct_addr:
                self._thread_cache = tps(int(thread_cache_struct_addr))
                return self._thread_cache

        # return the value of tcache if we have it cached
        if self._thread_caches.get(pwndbg.dbg.selected_thread().index()):
            return self._thread_caches[pwndbg.dbg.selected_thread().index()]

        arena = self.thread_arena
        if not arena:
            # arena doesn't be allocated yet, so there's no tcache
            return None

        if self.main_arena.next != self.main_arena.address or self.multithreaded:
            if self.prompt_for_brute_force_thread_cache_permission():
                tls_address = self.prompt_for_tls_address()
                if tls_address:
                    chunk_header_size = pwndbg.aglib.arch.ptrsize * 2
                    tcache_perthread_struct_size = self.tcache_perthread_struct.sizeof
                    lb, ub = arena.active_heap.start, arena.active_heap.end

                    def validator(guess: int) -> bool:
                        if guess < lb or guess >= ub:
                            return False
                        if not pwndbg.aglib.memory.is_readable_address(
                            guess - chunk_header_size
                        ) or not pwndbg.aglib.memory.is_readable_address(
                            guess + tcache_perthread_struct_size
                        ):
                            return False
                        chunk = Chunk(guess - chunk_header_size)
                        return chunk.real_size - chunk_header_size == tcache_perthread_struct_size

                    found = self.brute_force_tls_reference_in_got_section(
                        tls_address, validator
                    ) or self.brute_force_thread_local_variable_near_tls_base(
                        tls_address, validator
                    )
                    if found:
                        value, address = found
                        print(
                            message.notice(
                                f"Found possible tcache at {message.hint(hex(address))} with value: {message.hint(hex(value))}\n"
                            )
                        )
                        self._thread_cache = tps(value)
                        self._thread_caches[pwndbg.dbg.selected_thread().index()] = (
                            self._thread_cache
                        )
                        return self._thread_cache

            print(
                message.warn(
                    "Cannot find tcache, we assume it's at the beginning of the heap.\n"
                    "If you think this is wrong, please manually set it with `set tcache <address>`.\n"
                )
            )

        # TODO: The result might be wrong if the arena is being shared by multiple thread
        self._thread_cache = tps(arena.heaps[0].start + pwndbg.aglib.arch.ptrsize * 2)
        self._thread_caches[pwndbg.dbg.selected_thread().index()] = self._thread_cache

        return self._thread_cache

    @property
    def mp(self) -> pwndbg.aglib.heap.structs.CStruct2GDB:
        mp_via_config = int(str(pwndbg.config.mp), 0)
        mp_via_symbol = pwndbg.aglib.symbol.lookup_symbol_addr("mp_", prefer_static=True)
        if mp_via_config or mp_via_symbol:
            self._mp_addr = mp_via_symbol

        if not self._mp_addr:
            if self.is_statically_linked():
                section = pwndbg.aglib.proc.dump_elf_data_section()
                section_address = pwndbg.aglib.proc.get_section_address_by_name(".data")
            else:
                section = pwndbg.libc.section_by_name(".data")
                section_address = pwndbg.libc.section_address_by_name(".data")
            if section is not None and section_address:
                _, _, data = section

                # try to find the default mp_ struct in the .data section
                found = data.find(bytes(self.struct_module.DEFAULT_MP_))
                if found != -1:
                    self._mp_addr = section_address + found

        if pwndbg.aglib.memory.is_readable_address(self._mp_addr):
            mps = self.malloc_par
            self._mp = mps(self._mp_addr)
            return self._mp

        raise SymbolNotRecoveredError("mp_", "Could not find mp_ in the .data section.")

    @property
    def global_max_fast(self) -> int:
        global_max_fast_via_config = int(str(pwndbg.config.global_max_fast), 0)
        global_max_fast_via_symbol = pwndbg.aglib.symbol.lookup_symbol_addr(
            "global_max_fast", prefer_static=True
        )

        if global_max_fast_via_config or global_max_fast_via_symbol:
            self._global_max_fast_addr = global_max_fast_via_config or global_max_fast_via_symbol
            self._global_max_fast = pwndbg.aglib.memory.u(self._global_max_fast_addr)
            return self._global_max_fast

        # https://elixir.bootlin.com/glibc/glibc-2.37/source/malloc/malloc.c#L836
        # https://elixir.bootlin.com/glibc/glibc-2.37/source/malloc/malloc.c#L1773
        # https://elixir.bootlin.com/glibc/glibc-2.37/source/malloc/malloc.c#L1953
        default = (64 * self.size_sz // 4 + self.size_sz) & ~self.malloc_align_mask
        print(
            message.warn(
                f"global_max_fast symbol not found, using the default value: 0x{default:x}"
            )
        )
        print(
            message.warn(
                "Use `set global-max-fast <address>` to set the address of global_max_fast manually if needed."
            )
        )
        return default

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def heap_info(self) -> type[pwndbg.aglib.heap.structs.HeapInfo] | None:
        if not self.struct_module:
            return None
        return self.struct_module.HeapInfo

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_chunk(self) -> type[pwndbg.aglib.heap.structs.MallocChunk] | None:
        if not self.struct_module:
            return None
        return self.struct_module.MallocChunk

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_state(self) -> type[pwndbg.aglib.heap.structs.MallocState] | None:
        if not self.struct_module:
            return None
        return self.struct_module.MallocState

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def tcache_perthread_struct(
        self,
    ) -> type[pwndbg.aglib.heap.structs.TcachePerthreadStruct] | None:
        if not self.struct_module:
            return None
        return self.struct_module.TcachePerthreadStruct

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def tcache_entry(self) -> type[pwndbg.aglib.heap.structs.TcacheEntry] | None:
        if not self.struct_module:
            return None
        return self.struct_module.TcacheEntry

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def mallinfo(self) -> type[pwndbg.aglib.heap.structs.CStruct2GDB] | None:
        # TODO/FIXME: Currently, we don't need to create a new class for `struct mallinfo` because we never use it.
        raise NotImplementedError("`struct mallinfo` is not implemented yet.")

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def malloc_par(self) -> type[pwndbg.aglib.heap.structs.MallocPar] | None:
        if not self.struct_module:
            return None
        return self.struct_module.MallocPar

    def get_heap(self, addr: int) -> pwndbg.aglib.heap.structs.HeapInfo | None:
        """Find & read the heap_info struct belonging to the chunk at 'addr'."""
        hi = self.heap_info
        return hi(heap_for_ptr(addr))

    def get_tcache(
        self, tcache_addr: int | None = None
    ) -> pwndbg.aglib.heap.structs.TcachePerthreadStruct | None:
        if tcache_addr is None:
            return self.thread_cache

        tps = self.tcache_perthread_struct
        return tps(tcache_addr)

    def get_sbrk_heap_region(self) -> pwndbg.lib.memory.Page:
        """Return a Page object representing the sbrk heap region.
        Ensure the region's start address is aligned to SIZE_SZ * 2,
        which compensates for the presence of GLIBC_TUNABLES.
        This heuristic version requires some sanity checks and may raise SymbolNotRecoveredError
        if malloc's `mp_` struct can't be resolved.
        """
        # Initialize malloc's mp_ struct if necessary.
        if not self._mp_addr:
            try:
                self.mp
            except Exception:
                # Should only raise SymbolNotRecoveredError, but the heuristic heap implementation is still buggy so catch all exceptions for now.
                pass

        if self._mp_addr:
            if self.get_region(self.mp.get_field_address("sbrk_base")) and self.get_region(
                self.mp["sbrk_base"]
            ):
                assert isinstance(pwndbg.aglib.heap.current, GlibcMemoryAllocator)
                sbrk_base = pwndbg.lib.memory.align_up(
                    int(self.mp["sbrk_base"]), pwndbg.aglib.heap.current.size_sz * 2
                )

                sbrk_region = self.get_region(sbrk_base)
                if sbrk_region is None:
                    raise ValueError("mp_.sbrk_base is unmapped or points to unmapped memory.")
                sbrk_region.memsz = sbrk_region.end - sbrk_base
                sbrk_region.vaddr = sbrk_base

                return sbrk_region
            raise ValueError("mp_.sbrk_base is unmapped or points to unmapped memory.")
        raise SymbolNotRecoveredError("mp_", "Heuristic failed.")

    def is_initialized(self) -> bool:
        # TODO/FIXME: If main_arena['top'] is been modified to 0, this will not work.
        # try to use vmmap or main_arena.top to find the heap
        return any("[heap]" == x.objfile for x in pwndbg.aglib.vmmap.get()) or (
            self.can_be_resolved() and self.main_arena.top != 0
        )



"""
Project-wide function decorators.
"""

from __future__ import annotations

import functools
import traceback
from collections.abc import Callable
from typing import TypeVar

from typing_extensions import ParamSpec

import pwndbg
from pwndbg.color import message

first_prompt = False

P = ParamSpec("P")
T = TypeVar("T")
K = TypeVar("K")


def only_after_first_prompt(
    value_before: T | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T | None]]:
    """
    Decorator to prevent a function from running before the first prompt was displayed.
    The 'value_before' parameter can be used to specify the value that is
    returned if the function is called before the first prompt was displayed.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T | None]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T | None:
            if first_prompt:
                return func(*args, **kwargs)
            return value_before

        return wrapper

    return decorator


def suppress_errors(
    fallback: K = None, should_warn: bool = True
) -> Callable[[Callable[P, T]], Callable[P, T | K]]:
    """
    Decorator to make a function return a fallback value when it would otherwise error.
    The 'fallback' parameter can be used to specify the fallback value.
    If the 'should_warn' parameter is set, a warning will be printed whenever an error is suppressed.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T | K]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T | K:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if should_warn:
                    print(message.warn(f"Warning: Received an error in {func.__name__}: {e}"))
                    if pwndbg.config.exception_verbose:
                        traceback.print_exc()
                return fallback

        return wrapper

    return decorator



from __future__ import annotations

import argparse
import threading
from typing import Any

import pwndbg.aglib
import pwndbg.aglib.kernel
import pwndbg.aglib.kernel.slab
import pwndbg.aglib.symbol
import pwndbg.arguments
import pwndbg.color as color
import pwndbg.color.message as message
import pwndbg.commands.context
import pwndbg.dbg_mod
import pwndbg.lib.cache
from pwndbg.dbg_mod import BreakpointLocation
from pwndbg.dbg_mod import DebuggerType

parser = argparse.ArgumentParser(
    description="""
Trace kernel memory (SLUB and buddy) allocations and frees.

This command will execute `next` in the debugger, and print out all (de)allocations that happen until
the command finishes. As such this makes most sense to call when the PC is on a function call instruction.
Only (de)allocations triggered by the current function are considered (rather than other threads etc).

If neither `-s` nor `-b` are passed, both allocators are traced.
    """,
)
parser.add_argument(
    "-s", "--trace-slab", action="store_true", help="do only slab allocator tracing"
)
parser.add_argument(
    "-b", "--trace-buddy", action="store_true", help="do only buddy allocator tracing"
)
parser.add_argument("-v", "--verbose", action="store_true", help="print backtraces")
parser.add_argument(
    "-c",
    "--command",
    type=str,
    default="next",
    help="trace during the execution of this command",
)
parser.add_argument(
    "--all",
    action="store_true",
    help="display ALL memory allocations/frees regardless if they are triggered by the current function.",
)


class KmemTracepointsData:
    def __init__(self, verbose: bool, trace_all: bool) -> None:
        self.results: list[str] = []
        self.order = 0
        self.mutex = threading.RLock()
        self.verbose = verbose
        self.curr = None  # None means tracing all
        if not trace_all:
            # current frame only accounting for jumps
            if inf := pwndbg.dbg.selected_frame():
                if parent := inf.parent():
                    pc = parent.pc()
                    if symbol_name := pwndbg.aglib.symbol.resolve_addr(pc):
                        self.curr = symbol_name.split("+")[0]

            if self.curr is None:
                print(message.warn("Couldn't locate frame properly. Tracing --all."))

    def add_result(self, result: str) -> None:
        if not result:
            return
        with self.mutex:
            bt = pwndbg.commands.context.context_backtrace(False)
            if not self.curr or any(self.curr in line for line in bt):
                self.results.append(result)
                if self.verbose:
                    self.results += bt

    def _format_kmem_tracepoint_output(
        self, prefix: str, name: str, type: str, addr: int | None
    ) -> str:
        prefix = prefix.ljust(12, " ")
        if "FREE" in prefix:
            prefix = color.red(prefix)
        else:
            prefix = color.green(prefix)
        name = color.blue(name.ljust(20, " "))
        type = type.ljust(4, " ")
        return f"{prefix} {name} {type} @ {color.blue(hex(addr) if addr else str(addr))}"

    def format_slab_kmem_tracepoint_output(self, is_free: bool, objaddr: int | None) -> None:
        if is_free:
            prefix = "[SLAB FREE]"
        else:
            prefix = "[SLAB ALLOC]"
        cache = None
        if objaddr:
            _, cache = pwndbg.aglib.kernel.slab.find_containing_slab_cache(objaddr)
        result = message.warn(
            f"{prefix} invalid SLUB object @ {hex(objaddr) if objaddr else objaddr}"
        )

        if cache is not None:
            result = self._format_kmem_tracepoint_output(prefix, cache.name, "obj", objaddr)
        self.add_result(result)

    def format_page_kmem_tracepoint_output(
        self, is_free: bool, page: int | None, order: int
    ) -> None:
        if is_free:
            prefix = "[PAGE FREE]"
        else:
            prefix = "[PAGE ALLOC]"
        name = f"order-{order}"
        result = self._format_kmem_tracepoint_output(prefix, name, "page", page)
        if page is not None:
            physmap = pwndbg.aglib.kernel.page_to_virt(page)
            result += f" (physmap: {color.red(hex(physmap))})"
        self.add_result(result)


class KmemTracepoints:
    # exceptions to the default processing of (de)alloc funcs
    # the tuple is interpreted on a handler basis
    # e.g. `"kmem_cache_free": (1,)` means the second argument of kmem_cache_free is the address of freed object
    EXCEPTIONS: dict[str, tuple[Any, ...]] = {"kmem_cache_free": (1,)}

    def __init__(self) -> None:
        # try to capture the lowest possible level of exported functions in the (de)alloc chain
        # for example __alloc_pages_bulk calls __alloc_pages and only __alloc_pages is included
        # lists might not be complete
        # try to resolve all names, if does not exist, means it is not exported for that version
        # for the alloc functions, only the return value matters
        kmalloc_names = (  # (trys to) include all slab alloc functions for all v5.x and v6.x
            "__kmalloc",
            "__kmalloc_node",
            "__kmalloc_node_track_caller",
            "__kmalloc_track_caller",
            "__krealloc",
            "kmalloc_order",
            "kmalloc_order_trace",
            "kmem_cache_alloc",
            "kmem_cache_alloc_node",
            "kmem_cache_alloc_node_trace",
            "kmem_cache_alloc_trace",
            "kmem_cache_alloc_lru",
            "krealloc",
            "kmalloc_node_trace",
            "kmalloc_trace",
            "__kmalloc_node_noprof",
            "__kmalloc_noprof",
            "kmalloc_node_trace_noprof",
            "kmalloc_node_track_caller_noprof",
            "kmalloc_trace_noprof",
            "kmem_cache_alloc_lru_noprof",
            "kmem_cache_alloc_node_noprof",
            "kmem_cache_alloc_noprof",
            "krealloc_noprof",
            "__kmalloc_node_track_caller_noprof",
            "__kmalloc_cache_node_noprof",
            "__kmalloc_cache_noprof",
        )
        self.kallocs = self.resolve_names(kmalloc_names)
        kfree_names = (
            "kfree",
            "kfree_nolock",
            "kmem_cache_free",
        )
        self.kfrees = self.resolve_names(kfree_names)
        palloc_names = (  # all of those functions have the 2nd arg == order
            "__alloc_frozen_pages_noprof",
            "__alloc_pages",
            "__alloc_pages_nodemask",
            "alloc_pages_noprof",
        )
        self.pallocs = self.resolve_names(palloc_names)
        pfree_names = (  # page *, order
            "__free_pages",
        )
        self.pfrees = self.resolve_names(pfree_names)
        self.sps: list[pwndbg.dbg_mod.StopPoint] = []
        self.slab_tracepoints_enabled = True
        self.buddy_tracepoints_enabled = True

    def resolve_names(self, names: tuple[str, ...]) -> list[int]:
        result = []
        for name in names:
            addr = pwndbg.aglib.symbol.lookup_symbol_addr(name)
            if addr is None:
                continue
            result.append(addr)
        return result

    @staticmethod
    def _kalloc_handler() -> bool:
        assert pwndbg.aglib.regs.retval

        self = get_kmem_tracepoints()
        objaddr = pwndbg.aglib.regs.read_reg_uncached(pwndbg.aglib.regs.retval)
        self.data.format_slab_kmem_tracepoint_output(False, objaddr)
        return False

    @staticmethod
    def kalloc_handler(sp: pwndbg.dbg_mod.StopPoint) -> bool:
        pwndbg.dbg.selected_inferior().trace_ret(KmemTracepoints._kalloc_handler, True)
        return False

    @staticmethod
    def kfree_handler(sp: pwndbg.dbg_mod.StopPoint) -> bool:
        self = get_kmem_tracepoints()
        arg: int = 0
        if symbol_name := pwndbg.aglib.symbol.resolve_addr(pwndbg.aglib.regs.pc):
            symbol_name = symbol_name.split("+")[0]
            if symbol_name in KmemTracepoints.EXCEPTIONS:
                arg = KmemTracepoints.EXCEPTIONS[symbol_name][0]
        objaddr = pwndbg.arguments.argument(arg)
        self.data.format_slab_kmem_tracepoint_output(True, objaddr)
        return False

    @staticmethod
    def _palloc_handler() -> bool:
        assert pwndbg.aglib.regs.retval

        self = get_kmem_tracepoints()
        page = pwndbg.aglib.regs.read_reg_uncached(pwndbg.aglib.regs.retval)
        order = self.data.order
        self.data.format_page_kmem_tracepoint_output(False, page, order)
        return False

    @staticmethod
    def palloc_handler(sp: pwndbg.dbg_mod.StopPoint) -> bool:
        self = get_kmem_tracepoints()
        order = pwndbg.arguments.argument(1)
        pwndbg.dbg.selected_inferior().trace_ret(KmemTracepoints._palloc_handler, True)
        self.data.order = order
        return False

    @staticmethod
    def pfree_handler(sp: pwndbg.dbg_mod.StopPoint) -> bool:
        self = get_kmem_tracepoints()
        page = pwndbg.arguments.argument(0)
        order = pwndbg.arguments.argument(1)
        self.data.format_page_kmem_tracepoint_output(True, page, order)
        return False

    def register_breakpoints(self, verbose: bool, trace_all: bool) -> None:
        inf = pwndbg.dbg.selected_inferior()
        self.data = KmemTracepointsData(verbose, trace_all)
        self.data.results = []
        if self.slab_tracepoints_enabled:
            for kalloc in self.kallocs:
                bp = BreakpointLocation(kalloc)
                sp = inf.break_at(bp, KmemTracepoints.kalloc_handler, internal=True)
                self.sps.append(sp)
            for kfree in self.kfrees:
                bp = BreakpointLocation(kfree)
                sp = inf.break_at(bp, KmemTracepoints.kfree_handler, internal=True)
                self.sps.append(sp)
        if self.buddy_tracepoints_enabled:
            for palloc in self.pallocs:
                bp = BreakpointLocation(palloc)
                sp = inf.break_at(bp, KmemTracepoints.palloc_handler, internal=True)
                self.sps.append(sp)
            for pfree in self.pfrees:
                bp = BreakpointLocation(pfree)
                sp = inf.break_at(bp, KmemTracepoints.pfree_handler, internal=True)
                self.sps.append(sp)

    def remove_breakpoints(self) -> None:
        for sp in self.sps:
            sp.remove()
        self.sps = []
        self.slab_tracepoints_enabled = True
        self.buddy_tracepoints_enabled = True


@pwndbg.lib.cache.cache_until("objfile")
def get_kmem_tracepoints() -> KmemTracepoints:
    return KmemTracepoints()


@pwndbg.commands.Command(
    parser,
    category=pwndbg.commands.CommandCategory.KERNEL,
    notes="""
The `--all` flag may be helpful if you also want to trace frees scheduled with rcu or if the traced command
steps out of the current function. You may also find `-c finish` and `-c continue` useful.
""",
    # FIXME: The -c option
    #     default="next",
    # is not portable.
    only_debuggers={DebuggerType.GDB, DebuggerType.LLDB},
)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelSymbols
@pwndbg.commands.OnlyWhenPagingEnabled
def kmem_trace(trace_slab: bool, trace_buddy: bool, verbose: bool, command: str, all: bool) -> None:
    if pwndbg.aglib.regs.retval is None:
        print(
            message.error(
                "kmem-trace is not available on this architecture because the return value register is not defined."
            )
        )
        return

    tps = get_kmem_tracepoints()

    if not trace_slab and not trace_buddy:
        trace_slab = trace_buddy = True
    tps.slab_tracepoints_enabled = trace_slab
    tps.buddy_tracepoints_enabled = trace_buddy
    # We intentionally do not check `if trace_slab and trace_buddy` to allow for
    # commandline ergonomics.

    tps.register_breakpoints(verbose, all)
    print(message.success("Finished registering tracepoints."))

    old_val = pwndbg.config.context_backtrace_lines.value
    pwndbg.config.context_backtrace_lines.value = 1000  # enable full backtrace

    try:
        pwndbg.dbg.selected_inferior().runcmd(command)
    except Exception as e:
        tps.data.results.append(
            message.warn(f"running command `{command}` with kmem-trace failed with error: {str(e)}")
        )

    pwndbg.config.context_backtrace_lines.value = old_val  # restore
    pwndbg.commands.context.context()

    tps.remove_breakpoints()

    print("\n".join(tps.data.results))
    pwndbg.dbg.ctx_suspend_once()



from __future__ import annotations

import os
import struct
from typing import Any
from typing import NamedTuple

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.lib.cache
import pwndbg.lib.memory
from pwndbg.lib.arch import Platform

rw_flags = os.R_OK | os.W_OK
ro_flags = os.R_OK

# docs: https://github.com/pwndbg/pwndbg/issues/3261
_comm_start_page_rw = {
    "i386": 0xFFFF0000,
    "x86-64": 0x00007FFFFFE00000,
    "arm": 0xFFFF4000,
    "aarch64": 0x0000000FFFFFC000,
}

_comm_start_page_ro = {
    "arm": 0xFFFFC000,
    "aarch64": 0x0000000FFFFF4000,
}

_comm_max_size = 0xFFF


class CommPageField(NamedTuple):
    flags: int
    ctype: str
    name: str
    offset: int
    desc: str
    fmt: str

    def is_undocumented(self):
        return "UNDOCUMENTED" in self.name

    def is_unused(self):
        return "UNUSED" in self.name

    def is_readonly(self):
        return self.flags == ro_flags

    def real_addr(self) -> int:
        if self.is_readonly():
            page_start = _comm_start_page_ro[pwndbg.aglib.arch.name]
        else:
            page_start = _comm_start_page_rw[pwndbg.aglib.arch.name]
        return page_start + self.offset

    def real_size(self) -> int:
        return struct.calcsize(self.fmt)

    def unpack(self) -> Any:
        addr = self.real_addr()
        size = self.real_size()
        try:
            data = pwndbg.aglib.memory.read(addr, size)
            val = struct.unpack(self.fmt, data)[0]
        except pwndbg.dbg_mod.Error as e:
            val = f"<cannot read> ({e})"
        return val


# fmt: off
# docs: https://github.com/pwndbg/pwndbg/issues/3261
# docs: https://github.com/apple-oss-distributions/xnu/blob/e3723e1f17661b24996789d8afc084c0c3303b26/osfmk/arm/cpu_capabilities.h#L279-L384
_fields_arm = (
    CommPageField(rw_flags, "?", "COMM_PAGE_SIGNATURE", 0x000, "First few bytes contain a signature", "16s"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_CPU_CAPABILITIES64", 0x010, "CPU capabilities (64-bit)", "Q"),
    CommPageField(rw_flags, "?", "COMM_PAGE_UNUSED", 0x018, "Unused bytes", "6s"),
    CommPageField(rw_flags, "uint16", "COMM_PAGE_VERSION", 0x01E, "16-bit version number", "H"),
    CommPageField(rw_flags, "uint16", "COMM_PAGE_CPU_CAPABILITIES", 0x020, "CPU capabilities (32-bit)", "H"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_NCPUS", 0x022, "Number of configured CPUs", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_USER_PAGE_SHIFT_32_LEGACY", 0x024, "VM page shift for 32-bit processes (legacy)", "B"),
    CommPageField(ro_flags, "uint8", "COMM_PAGE_USER_PAGE_SHIFT_32", 0x024, "VM page shift for 32-bit processes", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_USER_PAGE_SHIFT_64_LEGACY", 0x025, "VM page shift for 64-bit processes (legacy)", "B"),
    CommPageField(ro_flags, "uint8", "COMM_PAGE_USER_PAGE_SHIFT_64", 0x025, "VM page shift for 64-bit processes", "B"),
    CommPageField(rw_flags, "uint16", "COMM_PAGE_CACHE_LINESIZE", 0x026, "Cache line size", "H"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_UNUSED4", 0x028, "Unused (was scheduler generation number)", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_UNUSED3", 0x02C, "Unused (was max spin count for mutexes)", "I"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_CPU_CLUSTERS", 0x02F, "Number of CPU clusters", "B"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_MEMORY_PRESSURE", 0x030, "Copy of VM memory pressure", "I"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_ACTIVE_CPUS", 0x034, "Number of active CPUs", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_PHYSICAL_CPUS", 0x035, "Number of physical CPUs", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_LOGICAL_CPUS", 0x036, "Number of logical CPUs", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_KERNEL_PAGE_SHIFT_LEGACY", 0x037, "Kernel VM page shift (legacy)", "B"),
    CommPageField(ro_flags, "uint8", "COMM_PAGE_KERNEL_PAGE_SHIFT", 0x037, "Kernel VM page shift", "B"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_MEMORY_SIZE", 0x038, "Maximum memory size", "Q"),
    CommPageField(rw_flags, "struct commpage_timeofday_data_t", "COMM_PAGE_TIMEOFDAY_DATA", 0x040, "Time-of-day data, gettimeofday() (legacy, 40 bytes)", "40s"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_CPUFAMILY", 0x080, "CPU family used by memcpy() resolver", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_DEV_FIRM_LEGACY", 0x084, "handle on PE_i_can_has_debugger (legacy)", "I"),
    CommPageField(ro_flags, "uint32", "COMM_PAGE_DEV_FIRM", 0x084, "handle on PE_i_can_has_debugger", "I"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_TIMEBASE_OFFSET", 0x088, "Timebase offset for mach_absolute_time()", "Q"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_USER_TIMEBASE", 0x090, "Is userspace mach_absolute_time supported", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_CONT_HWCLOCK", 0x091, "Always-on hardware clock presence for mach_continuous_time()", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_DTRACE_DOF_ENABLED", 0x092, "DTrace DOF enable flag", "B"),
    CommPageField(rw_flags, "?", "COMM_PAGE_UNUSED0", 0x093, "Unused bytes", "5s"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_CONT_TIMEBASE", 0x098, "Base for mach_continuous_time() relative to mach_absolute_time()", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_BOOTTIME_USEC", 0x0A0, "boottime in microseconds", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_CONT_HW_TIMEBASE", 0x0A8, "Base for mach_continuous_time() relative to hardware counter", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_APPROX_TIME", 0x0C0, "Last known mach_absolute_time()", "Q"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_APPROX_TIME_SUPPORTED", 0x0C8, "Is mach_approximate_time supported", "B"),
    CommPageField(rw_flags, "?", "COMM_PAGE_UNUSED1", 0x0D9, "Unused (padding to align cacheline)", "39s"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_KDEBUG_ENABLE", 0x100, "Kdebug status bits exported to userspace", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_ATM_DIAGNOSTIC_CONFIG", 0x104, "atm_diagnostic_config exported to userspace", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_MULTIUSER_CONFIG", 0x108, "multiuser_config exported to userspace", "I"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_JIT_WRITE_PROTECT", 0x10c, "Is jit_write_protect_supported_np", "B"),
    CommPageField(rw_flags, "struct new_commpage_timeofday_data_t", "COMM_PAGE_NEWTIMEOFDAY_DATA", 0x120, "gettimeofday(), struct new_commpage_timeofday_data_t (40 bytes)", "40s"),
    CommPageField(rw_flags, "struct bt_params", "COMM_PAGE_REMOTETIME_PARAMS", 0x148, "mach_bridge_remote_time(), struct bt_params (24 bytes)", "24s"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_DYLD_FLAGS", 0x160, "Dyld system flags kern.dyld_system_flags exported to userspace", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_CPU_QUIESCENT_COUNTER", 0x180, "CPU quiescent counter", "Q"),
    CommPageField(rw_flags, "?", "COMM_PAGE_CUSTOM2", 0x188, "Unused CPU quiescent counter (120 bytes reserved)", "120s"),
    CommPageField(rw_flags, "?", "COMM_PAGE_CPU_TO_CLUSTER", 0x200, "CPU ID to cluster ID mappings (256 bytes reserved)", "256s"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_VALUE", 0x320, "Random target value for security bounty", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_ADDRESS", 0x328, "Random target address for security bounty", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_KERN_VALUE", 0x330, "Random kernel value for security bounty", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_KERN_ADDRESS", 0x338, "Random kernel target address for security bounty", "Q"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_APT_MSG_POLICY", 0x340, "APT message policy APT_MSG", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_APT_ACTIVE", 0x341, "APT active status (infrequently mutated)", "B"),
)

# docs: https://github.com/pwndbg/pwndbg/issues/3261
# docs: https://github.com/apple-oss-distributions/xnu/blob/e3723e1f17661b24996789d8afc084c0c3303b26/osfmk/i386/cpu_capabilities.h#L185-L248
_fields_x86 = (
    CommPageField(rw_flags, "?", "COMM_PAGE_SIGNATURE", 0x000, "First 16 bytes contain a signature", "16s"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_CPU_CAPABILITIES64", 0x010, "CPU capabilities (64-bit)", "Q"),
    CommPageField(rw_flags, "?", "COMM_PAGE_UNUSED", 0x018, "Unused bytes", "6s"),
    CommPageField(rw_flags, "uint16", "COMM_PAGE_VERSION", 0x01E, "16-bit version number", "H"),
    CommPageField(rw_flags, "uint16", "COMM_PAGE_CPU_CAPABILITIES", 0x020, "CPU capabilities (32-bit, retained for compatibility)", "H"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_NCPUS", 0x022, "Number of configured CPUs (hw.logicalcpu at boot time)", "B"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_UNUSED0", 0x024, "Unused bytes (previously reserved for cpu_capabilities expansion)", "2s"),
    CommPageField(rw_flags, "uint16", "COMM_PAGE_CACHE_LINESIZE", 0x026, "Cache line size", "H"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_UNUSED4", 0x028, "Unused (was scheduler generation number)", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_MEMORY_PRESSURE", 0x02C, "Copy of VM memory pressure", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_UNUSED3", 0x030, "Unused (was max spin count for mutexes)", "I"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_ACTIVE_CPUS", 0x034, "Number of active CPUs", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_PHYSICAL_CPUS", 0x035, "Number of physical CPUs", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_LOGICAL_CPUS", 0x036, "Number of logical CPUs", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_CPU_CLUSTERS", 0x037, "Number of CPU clusters", "B"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_MEMORY_SIZE", 0x038, "Maximum memory size", "Q"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_CPUFAMILY", 0x040, "CPU family (hw.cpufamily, x86)", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_KDEBUG_ENABLE", 0x044, "Kdebug enable exported to userspace", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_ATM_DIAGNOSTIC_CONFIG", 0x048, "ATM diagnostic configuration exported to userspace", "I"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_DTRACE_DOF_ENABLED", 0x04C, "DTrace DOF enable flag", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_KERNEL_PAGE_SHIFT", 0x04D, "Kernel VM page shift (version >=14)", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_USER_PAGE_SHIFT_64", 0x04E, "User VM page shift (version >=14)", "B"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_UNUSED2", 0x04F, "Unused", "B"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_TIME_DATA_START", 0x050, "Base for time-related offsets", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_NT_TSC_BASE", 0x050, "Nanotime TSC base", "Q"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_NT_SCALE", 0x058, "Nanotime scale", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_NT_SHIFT", 0x05C, "Nanotime shift", "I"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_NT_NS_BASE", 0x060, "Nanotime nanosecond base", "Q"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_NT_GENERATION", 0x068, "Nanotime generation counter", "I"),
    CommPageField(rw_flags, "uint32", "COMM_PAGE_GTOD_GENERATION", 0x06C, "Gettimeofday generation counter", "I"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_GTOD_NS_BASE", 0x070, "Gettimeofday nanosecond base", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_GTOD_SEC_BASE", 0x078, "Gettimeofday second base", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_APPROX_TIME", 0x080, "Used by mach_approximate_time()", "Q"),
    CommPageField(rw_flags, "uint8", "COMM_PAGE_APPROX_TIME_SUPPORTED", 0x088, "Flag if mach_approximate_time() is supported", "B"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_CONT_TIMEBASE", 0x0C0, "Base for mach_continuous_time()", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_BOOTTIME_USEC", 0x0C8, "Boot time in microseconds", "Q"),
    CommPageField(rw_flags, "struct new_commpage_timeofday_data_t", "COMM_PAGE_NEWTIMEOFDAY_DATA", 0x0D0, "New time-of-day data (40 bytes)", "40s"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_DYLD_FLAGS", 0x100, "Dyld system flags exported to userspace", "Q"),
    CommPageField(rw_flags, "?", "COMM_PAGE_CPU_TO_CLUSTER", 0x108, "CPU ID to cluster ID mappings (256 bytes reserved)", "256s"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_VALUE", 0x320, "Random target value for security bounty", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_ADDRESS", 0x328, "Random target address for security bounty", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_KERN_VALUE", 0x330, "Random kernel value for security bounty", "Q"),
    CommPageField(rw_flags, "uint64", "COMM_PAGE_ASB_TARGET_KERN_ADDRESS", 0x338, "Random kernel target address for security bounty", "Q"),
)
# fmt: on


def _comm_fix_gaps(fields: tuple[CommPageField, ...]) -> tuple[CommPageField, ...]:
    fixed_fields = []
    prev_endaddr = 0
    for comm in fields:
        if comm.is_readonly():
            fixed_fields.append(comm)
            continue

        curr_addr = comm.offset
        if (gap_size := (curr_addr - prev_endaddr)) > 0:
            gap_addr = prev_endaddr
            fixed_fields.append(
                CommPageField(
                    rw_flags,
                    f"? uint{gap_size * 8}",
                    "GAP UNDOCUMENTED",
                    gap_addr,
                    "",
                    f"{gap_size}s",
                )
            )

        fixed_fields.append(comm)
        prev_endaddr = curr_addr + comm.real_size()

    if (gap_size := (_comm_max_size - prev_endaddr)) > 0:
        gap_addr = prev_endaddr
        fixed_fields.append(
            CommPageField(
                rw_flags, f"? uint{gap_size * 8}", "GAP UNDOCUMENTED", gap_addr, "", f"{gap_size}s"
            )
        )

    return tuple(fixed_fields)


_comm_fields = {
    "i386": _comm_fix_gaps(_fields_x86),
    "x86-64": _comm_fix_gaps(_fields_x86),
    "arm": _comm_fix_gaps(_fields_arm),
    "aarch64": _comm_fix_gaps(_fields_arm),
}


def get_commpage_fields() -> tuple[CommPageField, ...]:
    if pwndbg.aglib.arch.platform != Platform.DARWIN:
        return ()
    return _comm_fields.get(pwndbg.aglib.arch.name, ())


@pwndbg.lib.cache.cache_until("start")
def get_commpage_mappings() -> tuple[pwndbg.lib.memory.Page, ...]:
    if pwndbg.aglib.arch.platform != Platform.DARWIN:
        return ()

    start_rw = _comm_start_page_rw.get(pwndbg.aglib.arch.name, None)
    if start_rw is None or not pwndbg.aglib.memory.peek(start_rw):
        return ()

    ptrsize: int = pwndbg.aglib.arch.ptrsize

    start_ro = _comm_start_page_ro.get(pwndbg.aglib.arch.name, None)
    if start_ro is None or not pwndbg.aglib.memory.peek(start_ro):
        return (
            pwndbg.lib.memory.Page(start_rw, _comm_max_size, rw_flags, 0, ptrsize, "[commpage]"),
        )

    return (
        pwndbg.lib.memory.Page(start_ro, _comm_max_size, ro_flags, 0, ptrsize, "[commpage]"),
        pwndbg.lib.memory.Page(start_rw, _comm_max_size, rw_flags, 0, ptrsize, "[commpage]"),
    )



"""
For our REPL, we need to drive our own I/O with the process being debugged. This
module contains all the strategies we have for doing that.
"""

from __future__ import annotations

import os
import sys
import threading

import lldb
from typing_extensions import override

from pwndbg.dbg_mod.lldb.util import system_decode

if os.name == "posix":
    # We use select for files when not on POSIX. Additionally, we support pseudo
    # terminal devices under POSIX.
    import ctypes
    import select
    import signal
    import termios

    TERM_CONTROL_AVAILABLE = True
    SELECT_AVAILABLE = True
    PTY_AVAILABLE = True
else:
    # We sleep for a little bit when we don't have select.
    import time

    TERM_CONTROL_AVAILABLE = False
    SELECT_AVAILABLE = False
    PTY_AVAILABLE = False


# This is documented in Python's termios module, under tcgetattr, but, for some
# reason, there's no constant for it.
TC_LFLAG = 3


class OpportunisticTerminalControl:
    """
    Handles optional terminal control for a given file descriptor. Crucially,
    all the functions in this class should work regardless of whether terminal
    control is actually supported on not, but should do nothing in case it is
    not supported.
    """

    fd: int
    supported: bool

    def __init__(self, fd: int = -1):
        """
        Creates an opportunistic terminal control object for the given file
        descriptor. If no file descriptor is given, this class will try to open
        '/dev/tty', and use that.
        """
        if not TERM_CONTROL_AVAILABLE:
            # Preemptively disable this class if terminal control isn't possible
            # in this target, and do nothing else.
            self.supported = False
            return

        if fd == -1:
            try:
                fd = os.open("/dev/tty", os.O_RDWR)
            except (FileNotFoundError, PermissionError, OSError):
                # Flop and die.
                self.supported = False
                return

        self.fd = fd

        # Query for basic support for this file descriptor by querying its
        # attributes. If that fails, we assume the file descriptor we were
        # given does not support terminal control.
        try:
            termios.tcgetattr(fd)
            self.supported = True
        except termios.error:
            self.supported = False

    def _getattrbits(self, attri: int, mask: int) -> int:
        """
        Returns the result of applying the given bitmask to the given index in
        the array returned by termios.tcgetattr.
        """
        attr = termios.tcgetattr(self.fd)
        return attr[attri] & mask

    def _setattrbits(self, attri: int, mask: int, value: int) -> None:
        """
        Modifies the attribute integer at the given index in the array returned
        by termios.tcgetattr, then sets the terminal attributes to the resulting
        value.

        The new attribute integer will look like `(attr & ~mask) | value`.
        """
        attr = termios.tcgetattr(self.fd)
        attr[attri] = (attr[attri] & ~mask) | value
        termios.tcsetattr(self.fd, termios.TCSANOW, attr)

    def get_line_buffering(self) -> bool:
        """
        Gets the current state of line buffering for this terminal.
        """
        if not self.supported:
            return True
        return self._getattrbits(TC_LFLAG, termios.ICANON) != 0

    def set_line_buffering(self, enabled: bool) -> None:
        """
        Enables or disables line buffering for this terminal.
        """
        if not self.supported:
            return
        self._setattrbits(TC_LFLAG, termios.ICANON, termios.ICANON if enabled else 0)

    def get_echo(self) -> bool:
        """
        Gets the current state of echoing for this terminal.
        """
        if not self.supported:
            return True
        return self._getattrbits(TC_LFLAG, termios.ECHO) != 0

    def set_echo(self, enabled: bool) -> None:
        """
        Enables or disables echoing for this terminal.
        """
        if not self.supported:
            return
        self._setattrbits(TC_LFLAG, termios.ECHO, termios.ECHO if enabled else 0)


class IODriver:
    def stdio(self) -> tuple[str | None, str | None, str | None]:
        """
        The names for the stdin, stdout and stderr files, respectively. These
        will get passed as arguments to `SBTarget.Launch`
        """
        raise NotImplementedError()

    def start(self, process: lldb.Process) -> None:
        """
        Starts the handling of I/O by this driver on the given process.
        """
        raise NotImplementedError()

    def stop(self) -> None:
        """
        Stops the handling of I/O by this driver.
        """
        raise NotImplementedError()

    def on_output_event(self) -> None:
        """
        Hints that there might be data in either the standard output or the
        standard error streams. This should be called when an
        `eBroadcastBitSTDOUT` or `eBroadcastBitSTDERR` is encountered by the
        event loop.
        """
        raise NotImplementedError()

    def on_process_start(self, proc: lldb.SBProcess) -> None:
        """
        Allow the I/O driver an opportunity to change aspects of the process
        after it has been launched, but before it has started executing, if it
        so wishes.
        """
        raise NotImplementedError()

    def close(self) -> None:
        """
        Terminate this driver and release all resources associated with it.
        """
        raise NotImplementedError()


def get_io_driver() -> IODriver:
    """
    Instances a new IODriver using the best strategy available in the current
    system. Meaning a PTY on Unix and plain text on Windows.
    """
    if PTY_AVAILABLE:
        pty = make_pty()
        if pty is not None:
            worker, manager = pty
            return IODriverPseudoTerminal(worker=worker, manager=manager)
    return IODriverPlainText()


class IODriverPlainText(IODriver):
    """
    Plaintext-based I/O driver. It simply copies input from our standard input
    to the standard input of a given process, and copies output from the standard
    output of a given process to out standard output.
    """

    likely_output: threading.BoundedSemaphore
    in_thr: threading.Thread
    out_thr: threading.Thread
    stop_requested: threading.Event
    stop_fulfilled: threading.Semaphore
    start_requested: threading.Semaphore
    _closed: threading.Event
    _running: bool
    _stdout_nonblock_failed: bool
    _stderr_nonblock_failed: bool

    process: lldb.SBProcess

    def __init__(self):
        self.likely_output = threading.BoundedSemaphore(1)
        self.process = None
        self.stop_requested = threading.Event()
        self.start_requested = threading.BoundedSemaphore(2)
        self.stop_fulfilled = threading.BoundedSemaphore(2)
        self._closed = threading.Event()
        self._running = False
        self._stdout_nonblock_failed = False
        self._stderr_nonblock_failed = False

        assert self.start_requested.acquire()
        assert self.start_requested.acquire()
        assert self.stop_fulfilled.acquire()
        assert self.stop_fulfilled.acquire()

        self.in_thr = threading.Thread(target=self._handle_input)
        self.out_thr = threading.Thread(target=self._handle_output)
        self.in_thr.start()
        self.out_thr.start()

    @override
    def stdio(self) -> tuple[str | None, str | None, str | None]:
        return None, None, None

    def _handle_input(self):
        while not self._closed.is_set():
            if not self.start_requested.acquire(blocking=True, timeout=1):
                continue

            while not self.stop_requested.is_set():
                if SELECT_AVAILABLE:
                    select.select([sys.stdin], [], [], 0.2)

                try:
                    data = sys.stdin.read()
                    self.process.PutSTDIN(data)
                except (BlockingIOError, TypeError):
                    # We have to check for TypeError here too, as, even though you
                    # *can* set stdin into nonblocking mode, it doesn't handle it
                    # very gracefully.
                    #
                    # See https://github.com/python/cpython/issues/57531

                    # Ignore blocking errors, but wait for a little bit before
                    # trying again if we don't have select().
                    if not SELECT_AVAILABLE:
                        time.sleep(0.1)
            self.stop_fulfilled.release()

    def _handle_output(self):
        while not self._closed.is_set():
            if not self.start_requested.acquire(blocking=True, timeout=1):
                continue

            while not self.stop_requested.is_set():
                # Try to acquire the semaphore. This will not succeed until the next
                # process output event is received by the event loop.
                self.likely_output.acquire(blocking=True, timeout=0.2)

                # Don't actually stop ourselves, even if we can't acquire the
                # semaphore. LLDB can be a little lazy with the standard output
                # events, so we use the semaphore as way to respond much faster to
                # output than we otherwise would, but, even if we don't get an
                # event, we should still read the output, albeit at a slower pace.

                # Copy everything out to standard outputs.
                stdout = b""
                stderr = b""
                while True:
                    stdout += self.process.GetSTDOUT(1024).encode(sys.stdout.encoding)
                    stderr += self.process.GetSTDERR(1024).encode(sys.stderr.encoding)

                    if len(stdout) == 0 and len(stderr) == 0:
                        # Note that, even if we have pulled nothing new from LLDB,
                        # we still only exit the loop once we manage to push out
                        # both buffers in their entirety.
                        #
                        # This is consistent with the behavior of blocking on STDOUT
                        # and STDERR that we want, even if the underlying files are
                        # actually non-blocking.
                        break

                    try:
                        stdout = stdout[sys.stdout.buffer.write(stdout) :]
                        sys.stdout.buffer.flush()
                    except BlockingIOError as e:
                        # STDOUT is nonblocking at this point, and so writes may
                        # fail. We trim off however much we have managed to write
                        # from the buffer, and try again in the next iteration.
                        stdout = stdout[e.characters_written :]

                    try:
                        stderr = stderr[sys.stderr.buffer.write(stderr) :]
                        sys.stderr.buffer.flush()
                    except BlockingIOError as e:
                        # Same goes for STDERR as goes for STDOUT.
                        stderr = stderr[e.characters_written :]

                # Crucially, we don't release the semaphore here. Releasing is the
                # job of the on_output_event function.
            self.stop_fulfilled.release()

    @override
    def on_output_event(self) -> None:
        try:
            self.likely_output.release()
        except ValueError:
            # We haven't responded to the previous event yet. No matter, when
            # the output handler gets around to it, all the output from the
            # previous events will get processed.
            #
            # All that matters is that the output handler knows there's *some*
            # data to process.
            pass

    @override
    def on_process_start(self, proc: lldb.SBProcess) -> None:
        # We don't really want to do anything on process start.
        pass

    @override
    def start(self, process: lldb.Process) -> None:
        # Set up new threads and start processing I/O.
        assert self.process is None, "Multiple calls to start()"
        self.process = process
        self.stop_requested.clear()
        os.set_blocking(sys.stdin.fileno(), False)

        # Nonblocking output is NOT what we want, but in UNIX systems O_NONBLOCK
        # is set in the context of the so-called "open file description"[1][2],
        # rather than in the context of the file descriptor itself. So, these
        # systems will helpfully - and silently, of course - propagate a change
        # in blocking policy to all file descriptors that share the same open
        # file description - such as ones created through F_DUPFD or dup(2).
        #
        # Since, in general, we can't know how STDIN, STDOUT and STDERR are
        # related to each other ahead of time, and, more specifically, they
        # often share the exact same open file description, we have to be able
        # to gracefully handle the case in which setting O_NONBLOCK for STDIN
        # will also necessarily set it for STDOUT and STDERR.
        #
        # The strategy this class elects to use, then, is to explicitly set all
        # of them to the same blocking policy. While this doesn't solve the
        # issue, it at least makes it so that it's not as surprising as it would
        # be, otherwise. :)
        #
        # [1]: https://pubs.opengroup.org/onlinepubs/9799919799/
        # [2]: https://linux.die.net/man/2/fcntl
        try:
            os.set_blocking(sys.stdout.fileno(), False)
        except OSError:
            # It's not guaranteed that sys.stdout is actually backed by a file,
            # or that that file supports non-blocking operation. In fact, the
            # Pwndbg CLI itself supports swapping out output the output streams
            # as part of capturing command output.
            #
            # As such, we must also be able to gracefully handle this case.
            self._stdout_nonblock_failed = True

        try:
            os.set_blocking(sys.stderr.fileno(), False)
        except OSError:
            # Same as above.
            self._stderr_nonblock_failed = True

        self.start_requested.release(2)
        self._running = True

    @override
    def stop(self) -> None:
        # Politely ask for the I/O processors to stop, and wait until they have
        # stopped on their own terms.
        assert self._running, "Tried to stop an IODriverPlainText that is not running"
        self.stop_requested.set()
        self.stop_fulfilled.acquire(blocking=True)
        self.stop_fulfilled.acquire(blocking=True)

        os.set_blocking(sys.stdin.fileno(), True)

        # See start()
        try:
            os.set_blocking(sys.stdout.fileno(), True)
        except OSError:
            if not self._stdout_nonblock_failed:
                raise

        try:
            os.set_blocking(sys.stderr.fileno(), True)
        except OSError:
            if not self._stderr_nonblock_failed:
                raise

        self._stdout_nonblock_failed = False
        self._stderr_nonblock_failed = False
        self.process = None
        self._running = False

    @override
    def close(self) -> None:
        if self._running:
            self.stop()
        self._closed.set()
        self.in_thr.join()
        self.out_thr.join()


def make_pty() -> tuple[str, int] | None:
    """
    We need to make a pseudo-terminal ourselves if we want the process to handle
    naturally for the user. Returns a tuple with the path of the worker device
    and the file descriptor of the manager device if successful.
    """
    # These functions are only part of the Python Standard Library starting in
    # Python 3.13, so we can't do much better than this, unfortunately.
    try:
        if sys.platform == "linux":
            libc = ctypes.CDLL("libc.so.6")
            # O_RWDR | O_NOCTTY = 0x102
            pty = libc.posix_openpt(0x102)
        elif sys.platform == "darwin":
            libc = ctypes.CDLL("libSystem.B.dylib")
            # O_RWDR | O_NOCTTY = 0x131072
            pty = libc.posix_openpt(0x131072)
        else:
            # Not supported.
            return None
    except OSError:
        # Not supported.
        return None

    if pty <= 0:
        return None

    libc.ptsname.restype = ctypes.c_char_p
    name = libc.ptsname(pty)

    if libc.unlockpt(pty) != 0:
        libc.close(pty)
        return None

    try:
        name = system_decode(name)
    except UnicodeDecodeError:
        # The name of the terminal device is nonsensical to us, so we can't use
        # this PTY. Warn the user that getting the PTY has failed.
        print(f"warning: cannot interpret ptsname {name} as a string. not using a pseudo-terminal")
        return None

    return name, pty


LIVE_PSEUDO_TERMINAL_OBJECTS = False


class IODriverPseudoTerminal(IODriver):
    """
    pty-based I/O driver. Forwards input from standard input and has support for
    terminal width and height, and for terminal-based file operations on the
    program being debugged.
    """

    manager: int
    worker: str
    stop_requested: threading.Event
    input_buffer: bytes
    io_thread: threading.Thread
    process: lldb.SBProcess
    termcontrol: OpportunisticTerminalControl
    _stdout_nonblock_failed: bool
    _stderr_nonblock_failed: bool

    has_terminal_control: bool

    def __init__(self, manager: int, worker: str):
        assert PTY_AVAILABLE, (
            "IODriverPseudoTerminal should never be created unless PTY_AVAILABLE is set"
        )

        global LIVE_PSEUDO_TERMINAL_OBJECTS
        LIVE_PSEUDO_TERMINAL_OBJECTS = True

        self.manager = manager
        self.worker = worker

        # Try to set up our opportunistic control of the input terminal.
        self.termcontrol = OpportunisticTerminalControl()
        if not self.termcontrol.supported:
            print("warning: could not set up terminal control")

        # Put the manager in nonblocking mode.
        os.set_blocking(self.manager, False)

        # We could support querying the terminal size in older versions of Python,
        # too, but, for now, this should be good enough.
        #
        # TODO: Properly support terminal size queries in Python 3.10 and older.
        # Handle terminal resizes.
        if sys.version_info >= (3, 11):
            # The way we currently handle terminal resizing absolutely does not
            # support multipleinstances of IODriverPseudoTerminal, but we
            # shouldn't have more than one object live at a time anyway for the
            # REPL, so this is fine.
            try:
                terminal = open("/dev/tty", "rb")

                def handle_sigwinch(_sig, _frame):
                    # Tell vermin to ignore these. This block is
                    # gated behind Python 3.11.
                    size = termios.tcgetwinsize(terminal.fileno())  # novm
                    termios.tcsetwinsize(self.manager, size)  # novm

                signal.signal(signal.SIGWINCH, handle_sigwinch)
            except (FileNotFoundError, PermissionError, OSError):
                print(
                    "warning: no terminal device in /dev/tty, expect no support for terminal sizes"
                )

        self.stop_requested = threading.Event()
        self.input_buffer = b""
        self.process = None

        self._stdout_nonblock_failed = False
        self._stderr_nonblock_failed = False

    @override
    def stdio(self) -> tuple[str | None, str | None, str | None]:
        return self.worker, self.worker, self.worker

    def _handle_io(self):
        while not self.stop_requested.is_set():
            select.select([sys.stdin, self.manager], [self.manager], [], 0.2)

            try:
                while True:
                    data = os.read(sys.stdin.fileno(), 1024)
                    if len(data) == 0:
                        break
                    self.input_buffer += data
            except OSError:
                pass

            try:
                written = os.write(self.manager, self.input_buffer)
                self.input_buffer = self.input_buffer[written:]
            except OSError:
                pass

            try:
                while True:
                    data = os.read(self.manager, 1024)
                    if len(data) == 0:
                        break

                    while len(data) > 0:
                        try:
                            data = data[sys.stdout.buffer.write(data) :]
                            sys.stdout.buffer.flush()
                        except BlockingIOError as e:
                            data = data[e.characters_written :]

            except OSError:
                pass

    @override
    def start(self, process: lldb.Process) -> None:
        # Set up new threads and start processing I/O.
        assert self.process is None, "Multiple calls to start()"
        self.process = process
        self.stop_requested.clear()
        os.set_blocking(sys.stdin.fileno(), False)

        # Same reasoning as IODriverPlainText applies here.
        try:
            os.set_blocking(sys.stdout.fileno(), False)
        except OSError:
            self._stdout_nonblock_failed = True

        try:
            os.set_blocking(sys.stderr.fileno(), False)
        except OSError:
            self._stderr_nonblock_failed = True

        self.was_line_buffering = self.termcontrol.get_line_buffering()
        self.was_echoing = self.termcontrol.get_echo()

        self.termcontrol.set_line_buffering(False)
        self.termcontrol.set_echo(False)

        self.io_thread = threading.Thread(target=self._handle_io)
        self.io_thread.start()

    @override
    def stop(self) -> None:
        # Politely ask for the I/O processors to stop, and wait until they have
        # stopped on their own terms.
        self.stop_requested.set()
        self.io_thread.join()
        os.set_blocking(sys.stdin.fileno(), True)

        # Same reasoning as IODriverPlainText applies here.
        try:
            os.set_blocking(sys.stdout.fileno(), True)
        except OSError:
            if not self._stdout_nonblock_failed:
                raise

        try:
            os.set_blocking(sys.stderr.fileno(), True)
        except OSError:
            if not self._stderr_nonblock_failed:
                raise

        self.termcontrol.set_line_buffering(self.was_line_buffering)
        self.termcontrol.set_echo(self.was_echoing)

        self._stdout_nonblock_failed = False
        self._stderr_nonblock_failed = False

        self.process = None

    @override
    def on_output_event(self) -> None:
        # We drive our output ourselves.
        pass

    @override
    def on_process_start(self, proc: lldb.SBProcess) -> None:
        # Once we have `pwndbg.gdblib.shellcode` functioning, we could try to
        # attempt a "coup" of the controlling TTY for the process, here, so we
        # get to have the PTY we set up in this class as the main controller for
        # this process.
        #
        # TODO: Replace controlling PTY of the process once it is set up.
        pass

    @override
    def close(self) -> None:
        pass
